// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sampl\e Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		The CCache implements a lightweight cache to increase the speed in which
// 		XL can display realtime data. The cache is simply a two array of XLOPERS.
// 		This is the format in which xlSet accepts data. Each CRequest and CProperty
// 		instantiation has an offset into the cache asscoiated with it. Each
// 		properties and requests collection provides quick look ups using hash table.
// 		Once you have apointer to a CRequest or CProperty, you can access its value
// 		within the cachet htrough its asscoiated offset.    

// 		Though keeping the cache in this array allows for quick adding and updating
// 		of values, removing is slower. The X dimension of the array is the amount
// 		of properties in the largest request. 

#include "precomp.h"
#include "guid.h"
#include "dataobj.h"
#include "dataobjs.h"
#include "request.h"
#include "requests.h"
#include "prop.h"
#include "props.h"
#include "dirtycel.h"
#include "realtime.h"
#include "cache.h"

extern CLIPFORMAT g_cfWOSAXRT;                           	

LPCSTR GetPropertyNameFromSet(PROPERTYSET FAR *pPropertySet, DWORD dwPropertyID);

void AddInDebugOutput(LPEXCEPINFO pexcepinfo, LPTSTR pszFunction, LPTSTR pszDesc, SCODE scode);

/*
 * CCache::Create
 *
 * Purpose:
 *  Creates an instance of the cache and initializes it. Use of
 *	the class should never call new.
 *
 * Parameters:
 *  pDataObject    	Cache associated DataObject.
 *	bInitSize		Specifies whether create should call InitCacheSize
 *	ppCache			Pointer to return new Cache 
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::Create(CDataObject FAR *pDataObject, BOOL bInitSize
	, class CCache FAR* FAR* ppCache)
{   
	CCache FAR *pCache = NULL;

	TRY{
		if(FAILED(CheckPointer((LPVOID) ppCache, 0))){
    		AddInDebugOutput(NULL, TEXT("CCache::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppCache = NULL;
		
		pCache = new CCache();

		if(pCache == NULL){
			AddInDebugOutput(NULL, TEXT("CCache::Create")
					, IDS_DEBUG_COULDNOTALLOCCACHE, NULL); 
			MYRAISEEXCEPTION;
		}

		// Save pointer back to its DataObject
		pCache->m_pDataObject = pDataObject;
		
		if(bInitSize){
			if(FAILED(pCache->InitSize(0))){
				AddInDebugOutput(NULL, TEXT("CCache::Create")
						, IDS_DEBUG_COULDNOTALLOCCACHE, NULL); 
				MYRAISEEXCEPTION;
			}
		}
		
		// if valid pointer was sent, return new cache
		if(ppCache){
			*ppCache = pCache;
		}else{
			AddInDebugOutput(NULL, TEXT("CCache::Create")
					, IDS_DEBUG_INVALIDPARAMPOINTER, NULL); 
			MYRAISEEXCEPTION;
		}
			
		return S_OK;
	}

	EXCEPT{                                           
		// Clean-up
		
		if(pCache)
			delete pCache;
	
		AddInDebugOutput(NULL, TEXT("CCache::Create")
			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);
	}
}

/*
 * CCache constructor
 */

CCache::CCache()
{
	m_pxlCache = NULL;
	m_pDataObject = NULL;
	m_dwRequestsAlloced = 0;
}

/*
 * CCache destructor
 */

CCache::~CCache()
{
	if(SUCCEEDED(CheckPointer(m_pxlCache)))
		#ifdef WIN32
			VirtualFree(m_pxlCache, 0, MEM_RELEASE);
		#else
 			GlobalFreePtr(m_pxlCache);
		#endif	
}

/*
 * CCache::InitSize
 *
 * Purpose:
 *  Initializes the size of the cache.
 *
 * Parameters:
 *  dwInitSize    The size to make the cache
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::InitSize(DWORD dwInitSize){
	TRY{
		// Reserve maximum size of cache
		// If VBA developer specifies a MaxRequestCount then allocate that
		// amount right form the start. Else begin allocating in chunks 
		// as requests get added. 
		
		// As a feture, if VBA developer specifies the MaxRequestCount, 
		// Cache will not allow requests over that amount to be added.		

		if(m_dwRequestsAlloced == 0){		
			if(m_pDataObject->m_dwMaxRequests == 0){
				m_dwRequestsAlloced = REQUEST_BUFFER;	
			}else{
				m_dwRequestsAlloced = m_pDataObject->m_dwMaxRequests;	
			}
		}
    
    	// Depending on the OS allocate the cache    
		#ifdef _WIN32
			if(dwInitSize == 0){
				dwInitSize = m_dwRequestsAlloced * (m_pDataObject->m_dwMaxProperties - m_pDataObject->m_bDontIncludeNames) * sizeof(XLOPER);
			}
			
			m_pxlCache = (LPXLOPER) VirtualAlloc(NULL, 1000 * VIRTUALRESERVE, MEM_RESERVE, PAGE_READWRITE);

			if(m_pxlCache == NULL){
				AddInDebugOutput(NULL, TEXT("CCache::InitSize")
						, IDS_DEBUG_ALLOCFAILED, NULL); 
				MYRAISEEXCEPTION;
			}

			VirtualAlloc(m_pxlCache, dwInitSize, MEM_COMMIT, PAGE_READWRITE);
		#else
			if(dwInitSize == 0){
				dwInitSize = m_dwRequestsAlloced * (m_pDataObject->m_dwMaxProperties - m_pDataObject->m_bDontIncludeNames) * sizeof(XLOPER);
			}
		
			m_pxlCache = (LPXLOPER) GlobalAllocPtr(GHND, dwInitSize);
		#endif      

		if(m_pxlCache == NULL){
			AddInDebugOutput(NULL, TEXT("CCache::InitSize")
					, IDS_DEBUG_ALLOCFAILED, NULL); 
			MYRAISEEXCEPTION;
		}
	                                           
		// Init values to xlNil else xlSet won't work
		DWORD len = (DWORD)(m_dwRequestsAlloced * (m_pDataObject->m_dwMaxProperties - m_pDataObject->m_bDontIncludeNames));	                                       
		for(DWORD i=0 ; i < len  ; i++){
			m_pxlCache[i].xltype = xltypeNil;			
		}

		return S_OK;
	}

	EXCEPT{
		#ifdef _WIN32
			if(m_pxlCache){
				HANDLE hHeap = GetProcessHeap();
				HeapFree(hHeap, NULL, m_pxlCache);
			}
		#else
			if(m_pxlCache)
				GlobalFreePtr(m_pxlCache);
		#endif
		
   		AddInDebugOutput(NULL, TEXT("CCache::InitSize")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);
	}
}


/*
 * CCache::AddRequest
 *
 * Purpose:
 *  Adds new request to cache
 *
 * Parameters:
 *	bstrName	Name of Request to Add
 *	pRequest	Requests associated CRequest instantiation
 *	ppxlRequest	Pointer to return Requests cache offset
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::AddRequest(BSTR bstrName, CRequest FAR *pRequest
		, LPXLOPER FAR* ppxlRequest)
{   
	TRY{
		if(FAILED(CheckPointer((LPVOID) bstrName, 0))){
    		AddInDebugOutput(NULL, TEXT("CCache::AddRequest")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckPointer((LPVOID) pRequest, sizeof(CRequest)) )){
    		AddInDebugOutput(NULL, TEXT("CCache::AddRequest")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckPointer((LPVOID) ppxlRequest, 0))){
    		AddInDebugOutput(NULL, TEXT("CCache::AddRequest")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppxlRequest = NULL;
		
		// If there is no cache, alloc now	
		if(m_pxlCache == NULL){
			//Create XLSetCache
			if(FAILED(InitSize(FALSE))){
				AddInDebugOutput(NULL, TEXT("CCache::AddRequest")
						, IDS_DEBUG_COULDNOTALLOCCACHE, NULL); 
				MYRAISEEXCEPTION;
			}
		}
        
        // Figure where this request should go in cache.
		DWORD dwIndex = (m_pDataObject->m_pRequests->m_col.m_cElements-1) * (m_pDataObject->m_dwMaxProperties - m_pDataObject->m_bDontIncludeNames);
                        
		// If needed realloc cache and MaxRequests was not set by VBA developer        
        if(m_pDataObject->m_dwMaxRequests == 0 
        		&& (DWORD) m_pDataObject->m_pRequests->m_col.m_cElements > m_dwRequestsAlloced){
            
            // Set new size
            
            #ifdef _WIN32
				DWORD dwSize = REQUEST_BUFFER * (m_pDataObject->m_dwMaxProperties - m_pDataObject->m_bDontIncludeNames) * sizeof(XLOPER);
				DWORD dwOffSet = m_dwRequestsAlloced * (m_pDataObject->m_dwMaxProperties - m_pDataObject->m_bDontIncludeNames) * sizeof(XLOPER);

				VirtualAlloc(m_pxlCache
					, dwOffSet + dwSize, MEM_COMMIT, PAGE_READWRITE);
					
				// Set increment requests cache can hold
				m_dwRequestsAlloced += REQUEST_BUFFER;			                                               
            #else
				// Set increment requests cache can hold
				m_dwRequestsAlloced += REQUEST_BUFFER;			                                               

				DWORD dwSize = m_dwRequestsAlloced * (m_pDataObject->m_dwMaxProperties - m_pDataObject->m_bDontIncludeNames) * sizeof(XLOPER);
	
				m_pxlCache = (LPXLOPER) GlobalReAllocPtr(m_pxlCache
					, dwSize, GHND);	
        	#endif
        		
			if(m_pxlCache == NULL){
				AddInDebugOutput(NULL, TEXT("CCache::AddRequest")
						, IDS_DEBUG_ALLOCFAILED, NULL); 
				MYRAISEEXCEPTION;
			}        		
        }               
        
        if(ppxlRequest){
	 		m_pxlCache[dwIndex].xltype = xltypeNil;

	 		*ppxlRequest = &m_pxlCache[dwIndex];
		}else{
			AddInDebugOutput(NULL, TEXT("CCache::AddRequest")
					, IDS_DEBUG_INVALIDPARAMPOINTER, NULL); 
			MYRAISEEXCEPTION;
		}                    
		
		return S_OK;
	}

	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CCache::AddRequest")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);	                  
	}	
}

/*
 * CCache::RemoveRequest
 *
 * Purpose:
 *  Removes request from cache
 *
 * Parameters:
 *	lIndex	Request to remove
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::RemoveRequest(DWORD lIndex)
{                           
	return ResultFromScode(E_FAIL);                                      
} 


/*
 * CCache::AddProperty
 *
 * Purpose:
 *  Adds property to cache
 *
 * Parameters:
 *	pRequest		CRequest instantiation which this property contains
 *  ppxlProperty    Pointer to return property's cache offset
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::AddProperty(CRequest FAR *pRequest, LPXLOPER FAR *ppxlProperty)
{                           
	TRY{
		if(FAILED(CheckPointer((LPVOID) pRequest, sizeof(CRequest) ) )){
    		AddInDebugOutput(NULL, TEXT("CCache::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		if(FAILED(CheckPointer((LPVOID) ppxlProperty, 0))){
    		AddInDebugOutput(NULL, TEXT("CCache::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppxlProperty = NULL;
	
		// Calc property's offset into the cache                       
		DWORD dwOffset = pRequest->m_pProperties->m_col.m_cElements - (1 + pRequest->m_pDataObject->m_bDontIncludeNames);
	    
	    // Init the property's offset in the cache
		pRequest->m_pxlRequest[dwOffset].xltype = xltypeNum;
		pRequest->m_pxlRequest[dwOffset].val.num = 0;
        
        // If caller passes valid pointer
        if(ppxlProperty){
			*ppxlProperty = &(pRequest->m_pxlRequest[dwOffset]);
        }else{
			AddInDebugOutput(NULL, TEXT("CCache::AddProperty")
					, IDS_DEBUG_INVALIDPARAMPOINTER, NULL); 
        	MYRAISEEXCEPTION;
        }
        
        // Update the request's max column count 
        if(pRequest->m_pProperties->m_col.m_cElements > pRequest->m_pRequests->m_dwMaxColCount){
			pRequest->m_pRequests->m_dwMaxColCount = pRequest->m_pProperties->m_col.m_cElements;        	
        }
		
		return S_OK;	
	}

	EXCEPT{
  		AddInDebugOutput(NULL, TEXT("CCache::AddProperty")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);	                  
	}            
}

/*
 * CCache::AddProperty
 *
 * Purpose:
 *  Adds property to cache
 *
 * Parameters:
 *	pRequest		CRequest instantiation which this property contains
 *  lIndex			Property to remove
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::RemoveProperty(CRequest FAR* pRequest, DWORD lIndex)
{                           
	return ResultFromScode(E_FAIL);                                      
}                                    


/*
 * CCache::Fill
 *
 * Purpose:
 *  Fills entire cache by calling WOSA/XRT object's IDataObject::GetData
 *	this function is called when the VBA developer is attempting an
 *	DataObjects.GetData which is used to communicate to existing WOSA/XRT
 *  object.
 *
 * Parameters:
 *S ptg		Storage Medium return by IDataObject::GetData. Contains all 
 *				the data required to fill cache.
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::Fill(LPSTGMEDIUM pStg)
{
	LPBYTE					pData = NULL;
	LPMARKETDATA			pMarketData = NULL;
    LPUPDATEITEM			pUpdateItem = NULL;
    LPUPDATEITEMPROP		pItemValue = NULL;
    LPPROPERTYSET   		pPropertySet = NULL;
    DWORD 					i = 0;
	EASYBSTR FAR *			pPropSetVal = NULL;
	TCHAR 					buf[MAX_DIGITS]; // for converting numbers to strings
	CRequest FAR *			pRequest = NULL;
	CProperty FAR *			pProperty = NULL;
	CRequests FAR *			pRequests = m_pDataObject->m_pRequests;
	LPXLOPER 				pxRequest = NULL;
    CNode FAR * 			pNode = NULL;
	VARIANT FAR *			pv = NULL;    
	DWORD 					dwMaxProperties = 0;
	DWORD 					dwCurrentRequestCount = 0;
	VARIANT 				v;
    
	TRY{	
		// TODAY, add-in only supports TYMED_HGLOBAL
		if(pStg->tymed != TYMED_HGLOBAL){
			AddInDebugOutput(NULL, TEXT("CCache::Fill")
					, IDS_DEBUG_STGNOTSUPPORTED, NULL); 
			MYRAISEEXCEPTION;
		}

		//Lock main data stream...
		pData = (LPBYTE) GlobalLock(pStg->hGlobal);
 		pMarketData = (LPMARKETDATA) pData;
	                                  
		if(FAILED(CheckPointer(pMarketData, (UINT) min(pMarketData->cbStream, 64000) ))){
			AddInDebugOutput(NULL, TEXT("CCache::Fill")
					, TEXT("The MARKETDATA buffer is corrupt."), NULL); 
			MYRAISEEXCEPTION;
		}
	
		// Save header info	                                  
		m_pDataObject->m_dwStatus = pMarketData->dwStatus;
		m_pDataObject->m_dwSubStatus = pMarketData->dwSubStatus;
		m_pDataObject->m_dwVendorData1 = pMarketData->dwVendorData1;
		m_pDataObject->m_dwVendorData2 = pMarketData->dwVendorData2;
				
		pUpdateItem = (LPUPDATEITEM) (pData + pMarketData->dwItemOffset);

		if(FAILED(CheckPointer(pUpdateItem, sizeof(UPDATEITEM) ))){
			AddInDebugOutput(NULL, TEXT("CCache::Fill")
					, TEXT("The initial UPDATEITEM buffer is corrupt.")
					, NULL); 
			MYRAISEEXCEPTION;
		}

		// First, I need to calc the dimensions for the cache
		m_dwRequestsAlloced = 0;
		while( pUpdateItem != (LPUPDATEITEM)pData){
		
			if(dwMaxProperties < pUpdateItem->cProperties){
				dwMaxProperties = pUpdateItem->cProperties;			
			}                        
			
			pUpdateItem = (LPUPDATEITEM) (pData + pUpdateItem->dwNext);

			if(FAILED(CheckPointer(pUpdateItem, sizeof(UPDATEITEM) ))){
				AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, TEXT("An UPDATEITEM buffer is corrupt.")
						, NULL); 
				MYRAISEEXCEPTION;
			}
			
			m_dwRequestsAlloced++;
		}


		// Are the values valid		
		if(m_dwRequestsAlloced <= 0 || dwMaxProperties <= 0){
			AddInDebugOutput(NULL, TEXT("CCache::Fill")
					, TEXT("The cache is empty.")
					, NULL); 
			MYRAISEEXCEPTION;
		}

		// This must be set before CCache::InitSize
		m_pDataObject->m_dwMaxProperties = dwMaxProperties;
		// Initialize the cache's size
		// Allocate entire cache upfront
		if(FAILED(InitSize(m_dwRequestsAlloced * dwMaxProperties * sizeof(XLOPER)))){
			AddInDebugOutput(NULL, TEXT("CCache::Fill")
					, IDS_DEBUG_COULDNOTALLOCCACHE, NULL); 
			MYRAISEEXCEPTION;
		}
		
        // Set the DataObject MaxProperties
		m_pDataObject->put_MaxProperties((long) dwMaxProperties);
	    
	    // Get Data always includes a names
		m_pDataObject->put_IncludeRequestNames((long) TRUE);
		
		// Calc first UpdateItem
		pUpdateItem = (LPUPDATEITEM) (pData + pMarketData->dwItemOffset);

		if(FAILED(CheckPointer(pUpdateItem, sizeof(UPDATEITEM) ))){
			AddInDebugOutput(NULL, TEXT("CCache::Fill")
					, TEXT("An UPDATEITEM buffer is corrupt")
					, NULL); 
			MYRAISEEXCEPTION;
		}
        
		VariantInit(&v);

		// Turn off Single Cells updating.
		m_pDataObject->m_pDirtyCells->SetNotSingleCellRefreshable();	
		                              
		// Start looping throught the requests in buffer	                              
		while(pUpdateItem != (LPUPDATEITEM)pData){
			// Calc pointer to property set
	    	pPropertySet = (LPPROPERTYSET) (pData + pUpdateItem->dwPropertySet);

			if(FAILED(CheckPointer(pPropertySet, sizeof(PROPERTYSET) ))){
				AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, TEXT("A PROPERTYSET buffer is corrupt")
						, NULL); 
				MYRAISEEXCEPTION;
			}

            // Get pointer to Item Value structure
			pItemValue = (LPUPDATEITEMPROP) (pUpdateItem + 1);	

			if(FAILED(CheckPointer(pItemValue, sizeof(UPDATEITEMPROP) ))){
				AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, TEXT("An UPDATEITEMPROP buffer is corrupt")
						, NULL); 
				MYRAISEEXCEPTION;
			}

			// Get pointer to Item Value's data
			pv = (VARIANT FAR *) (((LPBYTE)pUpdateItem) + pItemValue->dwValueOffset);

			if(FAILED(CheckVariant(pv, FALSE))){
				AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, TEXT("A VARIANT is corrupt")
						, NULL); 
				MYRAISEEXCEPTION;
			}

			if(pv->vt == VT_BSTR){
				v.vt = VT_BSTR;

				#ifdef UNICODE
					LPCSTR psz = (LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD));
					int len = lstrlenA(psz);
					WCHAR	wc[256]; 

					memset(wc, 0, (size_t) 256);

					MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, psz
						, len, wc, 255);

					v.bstrVal = SysAllocString(wc);
				#else
					v.bstrVal = SysAllocString((LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD)));
				#endif
			}else{
				if(FAILED(VariantCopy(&v, pv))){
					AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, IDS_DEBUG_VARIANTCOPYFAILED, NULL); 
					MYRAISEEXCEPTION;
				}
			}

			// Get offset into cache
			pxRequest = (LPXLOPER) &(m_pxlCache[dwCurrentRequestCount 
					* (dwMaxProperties - m_pDataObject->m_bDontIncludeNames)]);
    
			if(FAILED(CheckPointer(pxRequest, sizeof(XLOPER) ))){
				AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, TEXT("Cache contains corrupt XLOPER.")
						, NULL); 
				MYRAISEEXCEPTION;
			}

			// Add new request to Requests collection
			if(FAILED(pRequests->Add(pxRequest, (BSTR) v.bstrVal
					, pUpdateItem->dwRequestID, &pRequest, &pNode))){

				AddInDebugOutput(NULL, TEXT("CCache::Fill")
					, TEXT("Cache contains corrupt XLOPER.")
					, NULL); 
				MYRAISEEXCEPTION;
			}
		    // Loop through properties...
			for(i=0;i<pUpdateItem->cProperties;i++){
				VARIANT vName;           
				
				// Get property's name
				VariantInit(&vName);
				vName.vt = VT_BSTR;

				vName.bstrVal = GetPropertyNameFromSet(pPropertySet, i);				
		
				if(FAILED(CheckPointer((LPVOID) vName.bstrVal, 0))){
					AddInDebugOutput(NULL, TEXT("CCache::Fill")
							, TEXT("A PROPERTYSET buffer is corrupt.")
							, NULL); 
					MYRAISEEXCEPTION;
				}
				
		        
		        // Add property to collection
				if(FAILED(pRequest->m_pProperties->Add(vName, &pProperty, &pNode))){
					VariantClear(&vName);
					AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, IDS_DEBUG_PROPSADDFAILED, NULL); 
					MYRAISEEXCEPTION;
				}
				
				// save property's offset into cache
				pProperty->m_pxlProperty = &(pRequest->m_pxlRequest[i]);
		
				if(FAILED(CheckPointer(pProperty->m_pxlProperty
						, sizeof(XLOPER) ))){

					AddInDebugOutput(NULL, TEXT("CCache::Fill")
							, TEXT("The cache contains a corrupt XLOPER.")
							, NULL); 
					MYRAISEEXCEPTION;
				}
				
				VariantClear(&vName);
			    
			    // Goto next ItemData        
				pItemValue = (LPUPDATEITEMPROP) (pUpdateItem + 1);	
		
				if(FAILED(CheckPointer(pItemValue, sizeof(UPDATEITEMPROP) ))){
					AddInDebugOutput(NULL, TEXT("CCache::Fill")
							, TEXT("An UPDATEITEMPROP is corrupt.")
							, NULL); 
					MYRAISEEXCEPTION;
				}

				pItemValue += i;
                                     
				////////////////////////////////////////
                // Update property ID hash
				pNode->m_dwID = pItemValue->dwPropertyID;

		   		#ifdef UNICODE
					_ultow(pNode->m_dwID, buf, 10);
				#else
					_ultoa(pNode->m_dwID, buf, 10);
				#endif

    			if(FAILED(pRequest->m_pProperties->m_col.m_phashIDs->Insert(buf, pNode))){
					AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, IDS_DEBUG_HASHTABLEINSERTFAIL, NULL); 
					MYRAISEEXCEPTION;
   				}
				//////////////////////////////////////////
                
				// Get pointer property's value
				pv = (VARIANT FAR *) (((LPBYTE)pUpdateItem) + pItemValue->dwValueOffset);
		
				if(FAILED(CheckVariant(pv, FALSE))){
					AddInDebugOutput(NULL, TEXT("CCache::Fill")
							, TEXT("A VARIANT is corrupt.")
							, NULL); 
					MYRAISEEXCEPTION;
				}

				// Copy value into cache
				if(pv->vt == VT_BSTR){
					// then if its a BSTR
					#ifdef UNICODE
						LPCSTR psz = (LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD));
						int len = lstrlenA(psz);
						WCHAR	wc[256]; 

						memset(wc, 0, (size_t) 256);

						MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, psz
							, len, wc, 255);

						BSTR bstr = SysAllocString(wc);
					#else
						BSTR bstr = SysAllocString((LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD)));
					#endif
						
					BSTRToXLOPER(pProperty->m_pxlProperty, bstr);

					SysFreeString(bstr);
				}else{
					// Everything else is easy
					VariantToXLOPER(pProperty->m_pxlProperty, pv);
				}

			}
		    
		    dwCurrentRequestCount++;

			float fcom = (float) 0.0;
			
			if(m_dwRequestsAlloced > 0){ // Just to make sure
				fcom = (float) ((float) dwCurrentRequestCount / (float) m_dwRequestsAlloced);
			}

			m_pDataObject->m_pRealTimeExt->m_pDataObjects->DisplayStatusWindow
				( TEXT("Intializing add-in's cache..."), fcom, TRUE );

			VariantClear(&v);
		    
			// Next, please
			pUpdateItem = (LPUPDATEITEM) (pData + pUpdateItem->dwNext);

			if(FAILED(CheckPointer(pUpdateItem, sizeof(UPDATEITEM) ))){
				AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, TEXT("An UPDATEITEMPROP is corrupt.")
						, NULL); 
				MYRAISEEXCEPTION;
			}
		}	
	    
	    // Unlock handle
    	GlobalUnlock(pStg->hGlobal);
    
		return S_OK; 
	}
	
	EXCEPT{                          
		// Today a single corruption prevents the dataobject from
		// being created.
	
	    GlobalUnlock(pStg->hGlobal);
		
   		AddInDebugOutput(NULL, TEXT("CCache::Fill")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);
	}
}
	
/*
 * CCache::Update
 *
 * Purpose:
 *	Updates cache with new data. The new data arrives in the add-in's
 *	IAdviseSink::OnDataChange
 *
 * Parameters:
 *	 pStg		Storage Medium passed in IAdviseSink::OnDataChange. Contains 
 *				all the data required to fill cache.
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::Update(LPSTGMEDIUM pStg)
{
	LPBYTE 				pData = NULL;
	LPMARKETDATA 		pMarketData = NULL;
    LPUPDATEITEM 		pUpdateItem = NULL;
    LPUPDATEITEMPROP	pItemValue = NULL;
    DWORD 				i = 0;	  
	DISPID 				dispid = -1;
	DISPPARAMS 			dispparams = {NULL, NULL, 0, 0};
	CRequest FAR		*pRequest = NULL;
    CProperty FAR		*pProperty = NULL;  
	VARIANT 			v;
	LPDISPATCH 			pDispatch = NULL;
	TCHAR 				buf[MAX_DIGITS]; // for converting numbers to strings
	CNode FAR *			pNode = NULL;
	BSTR				bstrPropertyName = NULL;

	TRY{
		// TODAY, add-in only supports TYMED_HGLOBAL
		if(pStg->tymed != TYMED_HGLOBAL){
			AddInDebugOutput(NULL, TEXT("CCache::Update")
					, IDS_DEBUG_STGNOTSUPPORTED, NULL); 
			MYRAISEEXCEPTION;
		}

		pData = (LPBYTE) GlobalLock(pStg->hGlobal);
		pMarketData = (LPMARKETDATA) pData;
	                                  
		if(FAILED(CheckPointer(pMarketData, (UINT) min(pMarketData->cbStream, 64000) ))){
			AddInDebugOutput(NULL, TEXT("CCache::Update")
					, TEXT("The MARKETDATA buffer is corrupt."), NULL); 
			MYRAISEEXCEPTION;
		}
	                                   
	    // Save header information
		m_pDataObject->m_dwStatus = pMarketData->dwStatus;
		m_pDataObject->m_dwSubStatus = pMarketData->dwSubStatus;
		m_pDataObject->m_dwVendorData1 = pMarketData->dwVendorData1;
		m_pDataObject->m_dwVendorData2 = pMarketData->dwVendorData2;
				                  
		// Get first UpdateItem				                  
		pUpdateItem = (LPUPDATEITEM) (pData + pMarketData->dwItemOffset);
		if(FAILED(CheckPointer(pUpdateItem, sizeof(UPDATEITEM) ))){
			AddInDebugOutput(NULL, TEXT("CCache::Update")
					, TEXT("The initial UPDATEITEM buffer is corrupt.")
					, NULL); 
			MYRAISEEXCEPTION;
		}
                                                        
		// Start Request Update Loop...	
		while(pUpdateItem != (LPUPDATEITEM)pData){
			// Finding request by ID is faster than by name so...
            //
			//1) Search for request in cache with this RequestID
			//
			if(FAILED(m_pDataObject->m_pRequests->m_col.FindObject(pUpdateItem->dwRequestID, &pDispatch, NULL) )){
				// If it's not there
				//
				// 2) Find request using Name property from Property Set
				//
				VariantInit(&v);			                                             
				LPUPDATEITEMPROP pItemValue = (LPUPDATEITEMPROP) (pUpdateItem + 1);	
				if(FAILED(CheckPointer(pItemValue, sizeof(UPDATEITEMPROP) ))){
					AddInDebugOutput(NULL, TEXT("CCache::Update")
						, TEXT("An UPDATEITEMPROP buffer is corrupt")
						, NULL); 
					MYRAISEEXCEPTION;
				}
				
				LPVARIANTARG pVar = (LPVARIANTARG) ((LPBYTE)pUpdateItem + pItemValue->dwValueOffset);

				if(pVar->vt == VT_BSTR){
					v.vt = VT_BSTR;
					#ifdef UNICODE
						LPCSTR psz = (LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD));
						int len = lstrlenA(psz);
						WCHAR	wc[256]; 

						memset(wc, 0, (size_t) 256);

						MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, psz
							, len, wc, 255);

						v.bstrVal = SysAllocString(wc);
					#else
						v.bstrVal = SysAllocString((LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD)));
					#endif
					
					if(FAILED(CheckPointer(v.bstrVal, 0))){
						AddInDebugOutput(NULL, TEXT("CCache::Update")
							, TEXT("Culd not extract propertyname from stream.")
							, NULL); 
						MYRAISEEXCEPTION;
					}
					
				}else{
					AddInDebugOutput(NULL, TEXT("CCache::Update")
							, IDS_DEBUG_NAMEPROPERTYNOTBSTR, NULL); 
					MYRAISEEXCEPTION;
           		}
						
                // This time try by request name
				if(FAILED(m_pDataObject->m_pRequests->m_col.FindObject(v.bstrVal, &pDispatch, NULL, &pNode))){
							
					/*****************************************************************************************
					*							Add optional code to do dymanic updating for 					     *
					*							NEW data not in automation collections!!!					 *
					*****************************************************************************************/
							
					AddInDebugOutput(NULL, TEXT("CCache::Update")
							, IDS_DEBUG_REQUESTNOTFOUNDINCACHE, NULL); 
					MYRAISEEXCEPTION;
				}else{                      
					// Internal QI is just a recast
					pDispatch->QueryInterface(IID_IRequest, (LPVOID FAR *) &pRequest);
					pDispatch->Release();
					pDispatch = NULL;
					
					//////////////////////////////////////////////			
					// At this point we should have a valid ID so ...
					// Save it for next time in hahs table
					pNode->m_dwID = pUpdateItem->dwRequestID;

				   	#ifdef UNICODE
						_ultow(pNode->m_dwID, buf, 10);
					#else
						_ultoa(pNode->m_dwID, buf, 10);
					#endif

    				if(FAILED(m_pDataObject->m_pRequests->m_col.m_phashIDs->Insert(buf, pNode))){
						AddInDebugOutput(NULL, TEXT("CCache::Update")
								, IDS_DEBUG_HASHTABLEINSERTFAIL, NULL); 
						MYRAISEEXCEPTION;
   					}
					///////////////////////////////////////////////
				}
			}else{                  
				// We have a valid REQUEST now recast it to a CRequest
			    // Internal QI is just a recast
				pDispatch->QueryInterface(IID_IRequest, (LPVOID FAR *) &pRequest);
				pDispatch->Release();
				pDispatch = NULL;
			}       	
		    
		    //
			// NOW, We must do the same for each property in loop
			//
			for(i=0;i<pUpdateItem->cProperties;i++){
				pItemValue = (LPUPDATEITEMPROP) (pUpdateItem + 1);	
				pItemValue += i;	

				if(FAILED(CheckPointer(pItemValue, sizeof(UPDATEITEMPROP) ))){
					AddInDebugOutput(NULL, TEXT("CCache::Update")
						, TEXT("An UPDATEITEMPROP buffer is corrupt")
						, NULL); 
					MYRAISEEXCEPTION;
				}
            
				//Search Rows for this RequestID                      
				if(FAILED(pRequest->m_pProperties->m_col.FindObject(pItemValue->dwPropertyID, &pDispatch, NULL))){
					//Find property using Name property from Property Set
					LPPROPERTYSET pPropertySet = (LPPROPERTYSET)(pData + pUpdateItem->dwPropertySet);

					bstrPropertyName = GetPropertyNameFromSet(pPropertySet, pItemValue->dwPropertyID);

					if(FAILED(CheckPointer((LPVOID) bstrPropertyName, 0))){
						AddInDebugOutput(NULL, TEXT("CCache::Fill")
							, IDS_DEBUG_COULDNOTPARSEPROPERTYNAME
							, NULL); 
						MYRAISEEXCEPTION;
					}
                               
					if(FAILED(pRequest->m_pProperties->m_col.FindObject(bstrPropertyName, &pDispatch, NULL, &pNode))){
						/******************************************************************************************
						*							Add Code to do dymanic aupdating for 					   	  *
						*							NEW data not in automation collections!!!					  *
						******************************************************************************************/

						AddInDebugOutput(NULL, TEXT("CCache::Update")
								, IDS_DEBUG_COULDNOTFINPROPERTY, NULL); 
						MYRAISEEXCEPTION;
					}else{
						pDispatch->QueryInterface(IID_IProperty, (LPVOID FAR *) &pProperty);
						pDispatch->Release();
						pDispatch = NULL;
			
						/*** At this point we should have a valid ID ***/
						// Save it for next time
						pNode->m_dwID = pItemValue->dwPropertyID;

						#ifdef UNICODE
							_ultow(pNode->m_dwID, buf, 10);
						#else
							_ultoa(pNode->m_dwID, buf, 10);
						#endif
	    
	    				if(FAILED(pRequest->m_pProperties->m_col.m_phashIDs->Insert(buf, pNode))){
							AddInDebugOutput(NULL, TEXT("CCache::Update")
									, IDS_DEBUG_HASHTABLEFINDFAIL, NULL); 
							MYRAISEEXCEPTION;
    					}
					}
				}else{
					// We have a valid PROPERTY now recast it to a CProperty
				    // Internal QI is just a recast

					pDispatch->QueryInterface(IID_IProperty, (LPVOID FAR *) &pProperty);
					pDispatch->Release();
					pDispatch = NULL;
				}		
                //
				// FINALLY, we can copy some data
				//
				
				VARIANT FAR *pv = (LPVARIANTARG) (((LPBYTE)pUpdateItem) + pItemValue->dwValueOffset);
				if(FAILED(CheckVariant(pv, FALSE))){
					AddInDebugOutput(NULL, TEXT("CCache::Update")
						, TEXT("A VARIANT is corrupt")
						, NULL); 
					MYRAISEEXCEPTION;
				}

				// Name must be handle specifically				
				if(!lstrcmpi(pProperty->m_bstrName, TEXT("Name"))){
					// Dont save Name if VBA developer specified not to
					if(!m_pDataObject->m_bDontIncludeNames){
						if(pv->vt == VT_BSTR){
							#ifdef UNICODE
								LPCSTR psz = (LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD));
								int len = lstrlenA(psz);
								WCHAR	wc[256]; 

								memset(wc, 0, (size_t) 256);

								MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, psz
									, len, wc, 255);

								BSTR bstr = SysAllocString(wc);
							#else
								BSTR bstr = SysAllocString((LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD)));
							#endif

							if(FAILED(BSTRToXLOPER(pProperty->m_pxlProperty, bstr))){
								AddInDebugOutput(NULL, TEXT("CCache::Update")
									, TEXT("Extracting value from stream failed.")
									, NULL); 
								SysFreeString(bstr);
								MYRAISEEXCEPTION;
							}
							
							SysFreeString(bstr);
						}else{
							AddInDebugOutput(NULL, TEXT("CCache::Update")
									, IDS_DEBUG_COULDNOTPARSEPROPERTYNAME, NULL); 
							MYRAISEEXCEPTION;
						}
                
						// DirtyCells allowes add-in to do updates per cells instead
						// of update per DataObject
						// Add dirty cell for potetntial single cell updating
						m_pDataObject->m_pDirtyCells->AddCell((WORD) pRequest->m_dwIndex, (WORD) (pProperty->m_dwIndex - m_pDataObject->m_bDontIncludeNames));
					}                   
				}else if(pv->vt == VT_BSTR){
					// then if its a BSTR
					#ifdef UNICODE
						LPCSTR psz = (LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD));
						int len = lstrlenA(psz);
						WCHAR	wc[256]; 

						memset(wc, 0, (size_t) 256);

						MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, psz
							, len, wc, 255);

						BSTR bstr = SysAllocString(wc);
					#else
						BSTR bstr = SysAllocString((LPCSTR) (((LPBYTE)pUpdateItem + pItemValue->dwValueOffset) + sizeof(VARIANT) + sizeof(DWORD)));
					#endif
						
					if(FAILED(BSTRToXLOPER(pProperty->m_pxlProperty, bstr))){
						AddInDebugOutput(NULL, TEXT("CCache::Update")
							, TEXT("Extracting value from stream failed.")
							, NULL); 
						SysFreeString(bstr);
						MYRAISEEXCEPTION;
					}

					SysFreeString(bstr);
                
					// DirtyCells allowes add-in to do updates per cells instead
					// of update per DataObject
					// Add dirty cell for potetntial single cell updating
					m_pDataObject->m_pDirtyCells->AddCell((WORD) pRequest->m_dwIndex, (WORD) (pProperty->m_dwIndex - m_pDataObject->m_bDontIncludeNames));
				}else{
					// Everything else is easy
					if(FAILED(VariantToXLOPER(pProperty->m_pxlProperty, pv))){
						AddInDebugOutput(NULL, TEXT("CCache::Update")
							, TEXT("Extracting value from stream failed.")
							, NULL); 
						MYRAISEEXCEPTION;
					}
                
					// DirtyCells allowes add-in to do updates per cells instead
					// of update per DataObject
					// Add dirty cell for potetntial single cell updating
					m_pDataObject->m_pDirtyCells->AddCell((WORD) pRequest->m_dwIndex, (WORD) (pProperty->m_dwIndex - m_pDataObject->m_bDontIncludeNames));
				}
				
				// Release from FindObject				
				pProperty->Release();
				pProperty = NULL;
			}
	    
			VariantClear(&v);

			//Release from FindObject				
			pRequest->Release();
			pRequest = NULL;

			// Goto next UpdateItem				
			pUpdateItem = (LPUPDATEITEM) (pData + pUpdateItem->dwNext);
			if(FAILED(CheckPointer(pUpdateItem, sizeof(UPDATEITEM) ))){
				AddInDebugOutput(NULL, TEXT("CCache::Fill")
						, TEXT("An UPDATEITEMPROP is corrupt.")
						, NULL); 
				MYRAISEEXCEPTION;
			}
		}	
	
		// Release the storage medium
		ReleaseStgMedium(pStg);
		
		// TYPEB UPDATE
		if(m_pDataObject->m_pUpdateRangeDispatch){
			if(FAILED(XLCalculateNow(m_pDataObject))){
				AddInDebugOutput(NULL, TEXT("CCache::Update")
						, TEXT("Attempt to recalc XLS failed.")
						, NULL); 
				MYRAISEEXCEPTION;
			}
		}

		return S_OK;
	}
	
	EXCEPT{
		ReleaseStgMedium(pStg);
		
		if(pProperty)
			pProperty->Release();
			
		if(pRequest)
			pRequest->Release();
			
		if(pDispatch)
			pDispatch->Release();
		
		VariantClear(&v);              
		                
		if(bstrPropertyName)
			SysFreeString(bstrPropertyName);

   		AddInDebugOutput(NULL, TEXT("CCache::Update")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);               
	}
}

/*
 * CCache::GetPropertyValue
 *
 * Purpose:
 *	Given request name and property name returns property's value from cache.
 *
 * Parameters:                             
 * 	ptrRequestName		Name of request which contains property
 *	ptrPropertyName		Name of property
 *	ppxlValue			Pointer to XLOPER to return value
 * 	
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CCache::GetPropertyValue(LPTSTR pstrRequestName
		, LPTSTR pstrPropertyName, XLOPER FAR* FAR* ppxlValue)
{
	LPDISPATCH pDispatch = NULL;
	CRequest FAR *pRequest = NULL;
	CProperty FAR *pProperty = NULL;

	TRY{
		if(FAILED(CheckPointer((LPVOID) pstrRequestName, 0))){
    		AddInDebugOutput(NULL, TEXT("CCache::GetPropertyValue")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckPointer((LPVOID) pstrPropertyName, 0))){
    		AddInDebugOutput(NULL, TEXT("CCache::GetPropertyValue")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckPointer((LPVOID) ppxlValue, 0))){
    		AddInDebugOutput(NULL, TEXT("CCache::GetPropertyValue")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppxlValue = NULL;
		
		// Find request
		if(FAILED(m_pDataObject->m_pRequests->m_col.FindObject(pstrRequestName
				, &pDispatch, NULL))){
			AddInDebugOutput(NULL, TEXT("CCache::GetPropertyValue")
					, IDS_DEBUG_REQUESTNOTFOUNDINCOLLECTION, NULL); 
			MYRAISEEXCEPTION;
		}
        
        // Recast pDispatch to pRequest
        // Internal QI is just a recast
		pDispatch->QueryInterface(IID_IRequest, (LPVOID FAR *) &pRequest);
        pDispatch->Release(); // Release from FindObject
        pDispatch = NULL;
        
		// Find property
		if(FAILED(pRequest->m_pProperties->m_col.FindObject(pstrPropertyName
				, &pDispatch, NULL))){
			AddInDebugOutput(NULL, TEXT("CCache::GetPropertyValue")
					, IDS_DEBUG_PROPNOTFOUNDINCOLLECTION, NULL); 
			MYRAISEEXCEPTION;
		}

        // Recast pDispatch to pProperty
        // Internal QI is just a recast
		pDispatch->QueryInterface(IID_IProperty, (LPVOID FAR *) &pProperty);
        pDispatch->Release(); // Release from FindObject
        pDispatch = NULL;
                                         
		// Return value                                         
        if(ppxlValue){
			*ppxlValue = pProperty->m_pxlProperty;        
		}
		
		pRequest->Release(); // Release from QI
		pRequest = NULL;
		
		pProperty->Release(); // Release from QI
        pProperty = NULL;
        
		return S_OK;               
	}	       
	
	EXCEPT{
		if(pRequest)
			pRequest->Release();
			
		if(pProperty)
			pProperty->Release();
			
		if(pDispatch)
			pDispatch->Release();
			
   		AddInDebugOutput(NULL, TEXT("CCache::GetPropertyValue")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		
		return ResultFromScode(E_FAIL);               
	}
}                                                                                

/*
 * CCache::GetPropertyNameFromSet
 *
 * Purpose:
 *	Given request name and property name returns property's value
 *
 * Parameters:                             
 * 	ptrRequestName	Name of request which contains property
 * 	ptrPropertyName	Name of property
 *	ppxlValue			Pointer to XLOPER to return value
 * 	
 * Return Value:
 *  HRESULT
 *
 */ 

BSTR CCache::GetPropertyNameFromSet(LPPROPERTYSET pPropertySet
			, DWORD dwPropertyID){
	DWORD i;
	EASYBSTR FAR *mbProperty;
	BSTR bstr;

	TRY{
		LPBYTE psz = (LPBYTE) (pPropertySet + 1);
    	
		for(i=0;i<=dwPropertyID;i++){
			mbProperty = (EASYBSTR FAR *) psz;

			if(FAILED(CheckPointer(mbProperty, sizeof(EASYBSTR) ))){
				return NULL;
			}
		
			psz += (mbProperty->dwNameLength + sizeof(DWORD));
		}
	
		#ifdef UNICODE
			WCHAR wc[256];
			int len = lstrlenA(mbProperty->szName);

			memset(wc, 0, (size_t) 256);

			MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED
			, mbProperty->szName, len, wc, 255);

			bstr = SysAllocString(wc);
		#else
			bstr = SysAllocString(mbProperty->szName);
		#endif

		return bstr;
	}

	#ifdef WIN32
	EXCEPT{
		return NULL;
	}
	#endif
}

