// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
// 		The CDataobject class implements the IUnknown and IDispatch interfaces.
// 		The IDispatch interface supersets all the WOSA/XRT DataObject methods 
// 		and properties. The other properties and methods are unique to to add-in
// 		and are defined throughout this module... 

#include "precomp.h"
#include "guid.h"
#include "framewrk.h"

#include "dataobj.h"
#include "requests.h"
#include "props.h"

#include "guid.h"
#include "realtime.h"
#include "dirtycel.h"
#include "iadvsink.h"
#include "cache.h"

#include <stdlib.h>
          
/*
 * CDataObject::Create
 *
 * Purpose:
 *  Creates an instance of the client object and initializes it. Use of
 *	the class should never call new.
 *
 * Parameters:
 *  ppDataObject    Returns Item object.
 *
 * Return Value:
 *  HRESULT
 *
 */ 

HRESULT CDataObject::Create(CDataObject FAR* FAR* ppDataObject
		, LPDISPATCH pXRTDispatch, BSTR bstrName, VARIANT vRange) 
{   
    CDataObject FAR* 	pDataObject = NULL;

	TRY{
		if(FAILED(CheckPointer((LPVOID) ppDataObject, 0))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppDataObject = NULL;
		
		if(FAILED(CheckDispatch(pXRTDispatch))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		if(FAILED(CheckPointer(bstrName, 0))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		pDataObject = new CDataObject();    
		
	    if (pDataObject == NULL){
			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_DATAOBJECTALLOCFAILED
					, NULL); 
			return ResultFromScode(E_FAIL);
    	}                                                                         
     
    	// Load type information for the collection from type library. CreateStdDispatch
    	// is not used to create a standard IDispatch implemenation because this
    	// collection implementation raises exceptions on ResultFromScode(E_FAIL)s.    
	    if (FAILED(LoadTypeInfo(&pDataObject->m_ptinfo, IID_IMyDataObject))){
			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_COULDNOTLOADTYPEINFO
					, NULL); 
			MYRAISEEXCEPTION;
    	}                                                                         
    
    	// Save parallel pointer to XRT Object
    	pDataObject->m_pXRTDispatch = pXRTDispatch;

		// Set recalc range if available                     
    	if(vRange.vt & VT_DISPATCH){
			if(SUCCEEDED(CheckVariant(&vRange, TRUE))){
	    		pDataObject->m_pUpdateRangeDispatch = vRange.pdispVal;
				pDataObject->m_pUpdateRangeDispatch->AddRef();
				pDataObject->m_pUpdateRangeDispatch->AddRef();
			}else{
				pDataObject->m_pUpdateRangeDispatch = NULL;
			}
		}
	
		// Call XRT Requests		
    	if(FAILED(pDataObject->XRTRequests(&pXRTDispatch))){
			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_CALLTOXRTDATAOBJREQUESTSFAILED
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
    	}
    
		// Create Requests Object    
    	if (FAILED(CRequests::Create(pXRTDispatch, &pDataObject->m_pRequests))){
			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_CREATEREQUESTSFAILED
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
	    }                                                                     
    
		// Save Parent    
    	pDataObject->m_pRequests->m_pDataObject = pDataObject; 

    	//Call XRT DataItems
	    if(FAILED(pDataObject->XRTDataItems(&pDataObject->m_pDataItemsXRTDispatch))){
			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_CALLTOXRTDATAOBJECTDATAITEMSFAILED
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
    	}

    	// Set Up Advise Sink
		pDataObject->m_pAdviseSink = new CImpIAdviseSink();
		
        if(pDataObject->m_pAdviseSink == NULL){
			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_CREATEIADVISESINKFAILED
					, DISP_E_EXCEPTION); 
			MYRAISEEXCEPTION;
        }
                                                                                  
		//Keep pointer back to parent                                                                                  
		pDataObject->m_pAdviseSink->m_pDataObject = pDataObject;	

		// Save clipboard format to use later                                           
		pDataObject->m_fte.cfFormat = g_cfWOSAXRT;
		pDataObject->m_fte.ptd = NULL;
		pDataObject->m_fte.dwAspect = DVASPECT_CONTENT;
 		pDataObject->m_fte.lindex = -1;
 		pDataObject->m_fte.tymed = TYMED_HGLOBAL;
        
        // IDataObject from WOSA/XRT object                                 
		if(FAILED(pDataObject->m_pXRTDispatch->QueryInterface(IID_IDataObject
				, (LPVOID FAR *) &(pDataObject->m_pXRTDataObject) ))){

			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_QIFORIDATAOBJECTFAILED
					, DISP_E_EXCEPTION); 

			MYRAISEEXCEPTION;
    	}

		// Set up async connection to WOSA/XRT object
   		if(FAILED(pDataObject->m_pXRTDataObject->DAdvise(
   				(LPFORMATETC) &pDataObject->m_fte, NULL
				, pDataObject->m_pAdviseSink, (DWORD FAR *) &pDataObject->m_dwAdviseConnection ) )){

			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_IDATAOBJECTDADVISEFAILED
					, DISP_E_EXCEPTION); 

			MYRAISEEXCEPTION;
	    }
    	                                             
		// Instantiate dirty cells cache    	                                          
    	if(FAILED(CDirtyCells::Create(&pDataObject->m_pDirtyCells))){

			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_DIRTYCELLSCREATEFAILED
					, DISP_E_EXCEPTION); 

			MYRAISEEXCEPTION;
    	}
		
		// Get hInstance for window creation
		XLOPER xRes;
		Excel(xlGetInst, (LPXLOPER) &xRes, 0);
		HINSTANCE hInstance = (HINSTANCE) xRes.val.w;
    	        
		Excel(xlGetHwnd, (LPXLOPER) &xRes, 0);
		HWND hwndXL = (HWND) xRes.val.w;

		#ifndef _WIN32
			SetMessageQueue(112);
    	#endif
   	    
		// Create window for status messages and 
		// in 16bit used to control update
		pDataObject->m_hWnd = CreateWindow(TEXT("XRTXLL_CLB_RC")
			, TEXT(""), WS_POPUP | WS_DLGFRAME
			, 0, 0, 10, 10
			, hwndXL, NULL, hInstance, (LPVOID) 0L);           

  		if (pDataObject->m_hWnd == NULL){

			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_CREATEWINDOWFAILED
					, DISP_E_EXCEPTION); 

			MYRAISEEXCEPTION;
		}                  
    	
    	// Position the Window, now		
		int sx = GetSystemMetrics(SM_CXSCREEN);
		int sy = GetSystemMetrics(SM_CYSCREEN);
        	
		RECT wr;
		GetWindowRect(pDataObject->m_hWnd, &wr);

		MoveWindow(pDataObject->m_hWnd, (sx-(wr.right-wr.left))/2, (sy-(wr.bottom-wr.top))/2
			, wr.right, wr.bottom, FALSE);
				
        #ifndef WIN32
			// Save pointer to Data Object for WndProc to use    
    		SetWindowLong(pDataObject->m_hWnd, 8, (LONG) pDataObject);
		#endif
		
		if(FAILED(CCache::Create(pDataObject, FALSE, &pDataObject->m_pCache))){

			AddInDebugOutput(&pDataObject->m_excepinfo
					, TEXT("CDataObject::Create")
					, IDS_DEBUG_CACHECREATEFAILED
					, DISP_E_EXCEPTION); 

			MYRAISEEXCEPTION;
		}
		
    	pDataObject->m_bstrName = SysAllocString(bstrName);

    	*ppDataObject = pDataObject;       
    
    	return S_OK;
	}
    
	EXCEPT{
   		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
   			, IDS_DEBUG_EXCEPTION, NULL); 

	    if(pDataObject)
	    	delete pDataObject;
	    
	    return ResultFromScode(E_FAIL);
	}
}

/*
 * CDataObject::CDataObject
 *
 * Purpose:
 *  Constructor for CDataObject object. Initializes members to NULL.
 *
 */                             
 
CDataObject::CDataObject()
{
    m_pDirtyCells = NULL;
	m_pUpdateRangeDispatch = NULL;
	m_bstrXLCallBackFunc = NULL;
	m_pRequests = NULL;
	m_pDataItemsXRTDispatch = NULL;
	m_pXRTDispatch = NULL;
	m_bstrName = NULL;       //Object Name
	m_hWnd = 0;
	m_dwAdviseConnection = 0;	
	m_pXRTDataObject = NULL;
	m_pAdviseSink = NULL;
	m_pRealTimeExt = NULL;                               
	m_ptinfo = NULL;
	m_cRef = 1;
   	m_bRaiseException = 0;
	m_dwMaxProperties = 0;
	m_dwMaxRequests = 0;
	m_pCache = NULL;
	m_bDontIncludeNames = 1;
	m_nUpdatePerCellPercentage = 1;
	m_bUpdateMaxRange = FALSE;
	m_bstrPassThroughFunc = NULL;
}
 

/*
 * CDataObject::~CDataObject
 *
 * Purpose:
 *  Destructor for CDataObject object. 
 *
 */

CDataObject::~CDataObject()
{                                                                        

	if(SUCCEEDED(CheckPointer(m_ptinfo)))
		m_ptinfo->Release();		

	if(SUCCEEDED(CheckPointer(m_pXRTDataObject)) && m_dwAdviseConnection)
		m_pXRTDataObject->DUnadvise(m_dwAdviseConnection);		
		
	if(SUCCEEDED(CheckDispatch(m_pUpdateRangeDispatch)))
		m_pUpdateRangeDispatch->Release;
       
	if(SUCCEEDED(CheckPointer(m_pRequests)))
		delete m_pRequests;    // this should delete my local. Release would work too
       
	if(SUCCEEDED(CheckDispatch(m_pDataItemsXRTDispatch)))
		m_pDataItemsXRTDispatch->Release();
    	
	if(SUCCEEDED(CheckPointer(m_bstrName)))
    	SysFreeString(m_bstrName);
    	         
	if(SUCCEEDED(CheckPointer(m_pAdviseSink)))
    	delete m_pAdviseSink;

	if(m_hWnd && IsWindow(m_hWnd))
		DestroyWindow(m_hWnd);    
		
	if(SUCCEEDED(CheckPointer(m_bstrXLCallBackFunc)))
		SysFreeString(m_bstrXLCallBackFunc);	                         
    	
	if(SUCCEEDED(CheckDispatch(m_pXRTDispatch)))
    	m_pXRTDispatch->Release();
    	
	if(SUCCEEDED(CheckPointer(m_pXRTDataObject)))
    	m_pXRTDataObject->Release();

	if(SUCCEEDED(CheckPointer(m_pDirtyCells)))
		delete m_pDirtyCells;			

	if(SUCCEEDED(CheckPointer(m_pCache)))
		delete m_pCache;
		
	if(SUCCEEDED(CheckPointer(m_bstrPassThroughFunc)))
		SysFreeString(m_bstrPassThroughFunc);					
}


/*
 * CDataObject::QueryInterface, AddRef, Release
 *
 * Purpose:
 *  Implements IUnknown::QueryInterface, AddRef, Release
 *
 */
 
STDMETHODIMP
CDataObject::QueryInterface(REFIID iid, void FAR* FAR* ppv) 
{   
    *ppv = NULL;
	
    if (iid == IID_IUnknown) 
		*ppv = this;  
    else if (iid == IID_IMyDataObject) 
		*ppv = this;    
    else if (iid == IID_IDispatch || iid == IID_DMyDataObject)
		*ppv = this;   
    else 
    	return ResultFromScode(E_NOINTERFACE); 

    AddRef();
    return S_OK;    
}


STDMETHODIMP_(ULONG)
CDataObject::AddRef(void)
{ 
   return ++m_cRef;
}


STDMETHODIMP_(ULONG)
CDataObject::Release(void)
{ 
    if(--m_cRef == 0){
		delete this;
		return 0;
    }
    return m_cRef;
} 


/*
 * CDataObject::GetTypeInfoCount
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfoCount.
 *
 */

STDMETHODIMP
CDataObject::GetTypeInfoCount(UINT FAR* pctinfo)
{
    *pctinfo = 1;
    return S_OK;
}

/*
 * CDataObject::GetTypeInfo
 *
 * Purpose:
 *  Implements IDispatch::GetTypeInfo. 
 *
 */

STDMETHODIMP
CDataObject::GetTypeInfo(
      UINT itinfo,
      LCID lcid,
      ITypeInfo FAR* FAR* pptinfo)
{    
    *pptinfo = NULL;
     
    if(itinfo != 0)
		return ResultFromScode(DISP_E_BADINDEX);
    
    m_ptinfo->AddRef(); 

    *pptinfo = m_ptinfo;
    
    return S_OK;
}

/*
 * CDataObject::GetIDsOfNames
 *
 * Purpose:
 *  Implements IDispatch::GetIDsOfNames.  The standard implementation, DispGetIDsOfNames,
 *  is used.
 *
 */

STDMETHODIMP 
CDataObject::GetIDsOfNames(
      REFIID riid,
      OLECHAR FAR* FAR* rgszNames,
      UINT cNames,
      LCID lcid,
      DISPID FAR* rgdispid)
{
    return DispGetIDsOfNames(m_ptinfo, rgszNames, cNames, rgdispid);
}

/*
 * CDataObject::Invoke
 *
 * Purpose:
 *  Implements IDispatch::Invoke.  The standard implementation, DispInvoke,
 *  is used. Properties and methods exposed by the collection object will
 *  set m_bRaiseException to raise an exception. 
 *
 *	In Excel RealTime Extensions version 1.00, I could not get automation
 *	exception exceptions to occurr consistently with descriptions and 
 *  wCodes. 
 *  
 */

STDMETHODIMP
CDataObject::Invoke(
      DISPID dispidMember,
      REFIID riid,
      LCID lcid,
      WORD wFlags,
      DISPPARAMS FAR* pdispparams,
      VARIANT FAR* pvarResult,
      EXCEPINFO FAR* pexcepinfo,
      UINT FAR* puArgErr)
{  
    HRESULT hr = S_OK;
     
    if (NULL == pexcepinfo)
		return ResultFromScode(E_INVALIDARG); 

   	m_bRaiseException = FALSE;
	m_bstrPassThroughFunc = NULL;	 
	
   	if(FAILED(DispInvoke(this, m_ptinfo,dispidMember, wFlags, pdispparams,
			pvarResult, pexcepinfo, puArgErr))){
	
		pexcepinfo->bstrSource = SysAllocString(g_szTitle); 
		pexcepinfo->bstrDescription = SysAllocString(TEXT("Internal invoke error. Type library may be corrupt.")); 
       	pexcepinfo->wCode = 1170;

		PrepareExcepInfo(pexcepinfo);

		return ResultFromScode(DISP_E_EXCEPTION); 
	}
	    
	if(m_bstrPassThroughFunc != NULL){
		DISPID xrtdispid = -1;

		if(FAILED(CheckDispatch(m_pXRTDispatch))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    			, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(m_pXRTDispatch->GetIDsOfNames(IID_NULL
			, (LPTSTR FAR *) &m_bstrPassThroughFunc, 1, lcid, &xrtdispid))){

			pexcepinfo->bstrSource = SysAllocString(g_szTitle); 
			pexcepinfo->bstrDescription = SysAllocString(TEXT("Dispatch call to bound XRT object failed.")); 
	       	pexcepinfo->wCode = 1171;

			PrepareExcepInfo(pexcepinfo);

			return ResultFromScode(DISP_E_EXCEPTION); 
		}

		SysFreeString(m_bstrPassThroughFunc);
		m_bstrPassThroughFunc = NULL;

		if(FAILED(m_pXRTDispatch->Invoke(xrtdispid, riid
			, lcid, wFlags, pdispparams, pvarResult
			, pexcepinfo , puArgErr))){

			pexcepinfo->bstrSource = SysAllocString(g_szTitle); 
			pexcepinfo->bstrDescription = SysAllocString(TEXT("Dispatch call to bound XRT object failed.")); 
	       	pexcepinfo->wCode = 1171;

			PrepareExcepInfo(pexcepinfo);
	
			return ResultFromScode(DISP_E_EXCEPTION); 
		}

		return S_OK;
	}

   	// Check if property accessor function or method called by DispInvoke wants to raise
   	// an exception. If so, return DISP_E_EXCEPTION after filling out the EXCEPINFO 
   	// structure. The scode and bstrDescription fields of EXCEPINFO are filled by the 
   	// property or method.    
   	if (m_bRaiseException){
		pexcepinfo->bstrDescription = m_excepinfo.bstrDescription;
		pexcepinfo->bstrSource = SysAllocString(g_szTitle);   
		pexcepinfo->wCode = m_excepinfo.wCode;   

		PrepareExcepInfo(pexcepinfo);

		m_bRaiseException = FALSE;
       	m_excepinfo.scode = NULL;
       	m_excepinfo.wCode = 0;

		return ResultFromScode(DISP_E_EXCEPTION); 
   	}else{
   		return hr;    
	}
}

/*
 * CDataObject::UpdateRange - Read/Write property
 *
 * Purpose:
 * 	If TypeB updating is being used, UpdateRange is an IDispatch pointer
 *	to the range to recalc.
 *
 * Parameters:
 *  newvalueDis ppatch to XL cell range to recalc on updates
 *
 * Return Value:
 *  HRESULT
 *
 */
  
STDMETHODIMP_(void)
CDataObject::put_UpdateRange(LPDISPATCH newvalue){
	// Release old one	
	if(SUCCEEDED(CheckDispatch(m_pUpdateRangeDispatch))){
	 	m_pUpdateRangeDispatch->Release();
	 	m_pUpdateRangeDispatch = NULL;
	}                
	
	// Keep new one
   	m_pUpdateRangeDispatch = newvalue;
    m_pUpdateRangeDispatch->AddRef();	

	return;	
}

STDMETHODIMP_(LPDISPATCH)
CDataObject::get_UpdateRange(void){
	if(FAILED(CheckDispatch(m_pUpdateRangeDispatch))){
		AddInDebugOutput(&m_excepinfo
				, TEXT("CDataObject::get_UpdateRange")
				, IDS_UPDATERANGEINVALID
				, DISP_E_EXCEPTION); 

		m_bRaiseException = TRUE;

		return NULL;
	}
  
	m_pUpdateRangeDispatch->AddRef();	
	
	return m_pUpdateRangeDispatch;	
}


/*
 * CDataObject::Requests - Read- Only property
 *
 * Purpose:
 *	Returns the DataObject's Requests collection	
 *
 */
  
STDMETHODIMP_(IDispatch *) 
CDataObject::get_Requests(void){                              
	LPDISPATCH pDispatch;

	// Recast to IDispatch
	// Internal QI is just a recast
	m_pRequests->QueryInterface(IID_IDispatch, (LPVOID FAR *) &pDispatch);

	return pDispatch;
}


/*
 * CDataObject::DataItems - Read-only property
 *
 * Purpose:
 *	Returns the DataObject's DataItems collection
 *  There is no reason to proxy the DataItems hierarchy so let VBA developer
 *	go directly to the WOSA/XRT object.	
 *
 * Return Value:
 *  LPDISPATCH
 *
 */

STDMETHODIMP_(IDispatch *) 
CDataObject::get_DataItems(void){
	m_pDataItemsXRTDispatch->AddRef();	                                    

	return m_pDataItemsXRTDispatch;
}
    
/*
 * CDataObject::Name - Read-only property
 *
 * Purpose:
 *	Returns the DataObject's Name. Add-in will overide the Name property
 *
 * Return Value:
 *  BSTR
 *
 */

STDMETHODIMP_(BSTR)
CDataObject::get_Name(void){
	return SysAllocString(m_bstrName);
}

/*
 * The follwoing properties and methods are standard to WOSA/XRT
 * Add-in simply passes functionality through to bound XRT object.
 * See WOSA/XRT Developers Toolkit for more information
 *
 * Add-in must still implement functions since mktyplib will generate
 * a header file with a pure virtual class to derive from. I rather not
 * edit the header file instead I implement each function.
 * 
 * CDataObject::UpdateFrequency 
 * CDataObject::Status - Read-only property
 * CDataObject::MostRecentlyOnly
 * CDataObject::Active
 * CDataObject::ColumnDelimiter
 * CDataObject::UserName
 * CDataObject::Password
 * CDataObject::WorkstationID
 * CDataObject::DataUserID
 * CDataObject::Activate
 * CDataObject::Copy
 * CDataObject::Close
 * CDataObject::Application
 * CDataObject::Visible
 * CDataObject::Save
 * CDataObject::SaveAs
 * CDataObject::Path
 * CDataObject::FullName
 * CDataObject::Title
 *
 * Purpose:
 *	Returns DataObjects UpdateFrequency
 *
 * Return Value:
 *  long
 *
 */

STDMETHODIMP_(long)
CDataObject::get_UpdateFrequency(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("UpdateFrequency"));

	return 0L;
}


STDMETHODIMP_(void)
CDataObject::put_UpdateFrequency(long newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("UpdateFrequency"));
    
	return;		
}


STDMETHODIMP_(long)
CDataObject::get_Status(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Status"));

	return NULL;		
}

STDMETHODIMP_(short)
CDataObject::get_MostRecentOnly(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("MostRecentlyOnly"));

	return NULL;		
}

STDMETHODIMP_(void)
CDataObject::put_MostRecentOnly(short newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("MostRecentlyOnly"));

	return;		
}

STDMETHODIMP_(short)
CDataObject::get_Active(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Active"));

	return NULL;		
}

STDMETHODIMP_(void)
CDataObject::put_Active(short newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Active"));

	return;		
}
                                                   
                                                   

STDMETHODIMP_(BSTR)
CDataObject::get_ColumnDelimiter(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("ColumnDelimiter"));

	return NULL;		
}

STDMETHODIMP_(void)
CDataObject::put_ColumnDelimiter(BSTR newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("ColumnDelimiter"));

	return;		
}


STDMETHODIMP_(BSTR)
CDataObject::get_Username(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Username"));

	return NULL;		
}

STDMETHODIMP_(void)
CDataObject::put_Username(BSTR newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Username"));

	return;		
}


STDMETHODIMP_(BSTR)
CDataObject::get_Password(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Password"));

	return NULL;		
}

STDMETHODIMP_(void)
CDataObject::put_Password(BSTR newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Password"));

	return;		
}

STDMETHODIMP_(VARIANT)
CDataObject::get_WorkstationID(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("WorkstationID"));

	VARIANT v;

	return v;		
}

STDMETHODIMP_(void)
CDataObject::put_WorkstationID(VARIANT newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("WorkstationID"));

	return;		
}

STDMETHODIMP_(VARIANT)
CDataObject::get_DataUserID(void){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("DataUserID"));

	VARIANT v;

	return v;		
}

STDMETHODIMP_(void)
CDataObject::put_DataUserID(VARIANT newval){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("DataUserID"));

	return;		
}

STDMETHODIMP_(void)
CDataObject::Activate(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Activate"));

	return;		
}

STDMETHODIMP_(void)
CDataObject::Copy(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Copy"));

	return;		
}

STDMETHODIMP_(VARIANT)
CDataObject::Close(VARIANT SaveChanges, VARIANT FileName){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Close"));

	VARIANT v;

	return v;		
}

STDMETHODIMP_(VARIANT)
CDataObject::Save(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Save"));

	VARIANT v;

	return v;		
}

STDMETHODIMP_(VARIANT)
CDataObject::SaveAs(VARIANT FileName){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("SaveAs"));

	VARIANT v;

	return v;		
}

STDMETHODIMP_(VARIANT)
CDataObject::get_Application(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Application"));

	VARIANT v;

	return v;		
}

STDMETHODIMP_(void)
CDataObject::put_Visible(short newvalue){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Visible"));

	return;
}

STDMETHODIMP_(short) 
CDataObject::get_Visible(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Visible"));

	return NULL;		
}

STDMETHODIMP_(short) 
CDataObject::get_Saved(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Saved"));

	return NULL;		
}


STDMETHODIMP_(BSTR) 
CDataObject::get_Path(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Path"));

	return NULL;		
}

STDMETHODIMP_(BSTR) 
CDataObject::get_FullName(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("FullName"));

	return NULL;		
}

STDMETHODIMP_(void) 
CDataObject::put_Title(BSTR newvalue){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Title"));

	return;		
}

STDMETHODIMP_(BSTR) 
CDataObject::get_Title(){
	// This property or method is an example of an XRT member which the
	// add-in does not need to be aware of. We can simply passes the 
	// call onto the XRT object. I do this in a centralized location:
	// ::Invoke.

	m_bstrPassThroughFunc = SysAllocString(TEXT("Title"));

	return NULL;		
}
                                                   
// ELIMNATE


// MY METHODS
    
/*
 * CProperty::Object - Read-only property
 *
 * Purpose:
 *	Returns WOSA/XRT object CProperty object. Used by VBA developer to call
 *	properties and methods which are specific to the WOSA/XRT object.
 *
 * Return Value:
 *  VARIANT
 *
 */

STDMETHODIMP_(IDispatch *)
CDataObject::get_Object(){
	if(FAILED(CheckDispatch(m_pXRTDispatch))){
    	AddInDebugOutput(NULL, TEXT("CDataObject::Create")
    		, IDS_DEBUG_BADPOINTERTOXRTOBJECT, NULL); 

		m_bRaiseException = TRUE;
		return NULL;
	}

	m_pXRTDispatch->AddRef();	
                   
	return m_pXRTDispatch;                     
}

/*
 * CDataObject::GetData		
 * Method
 *
 * Purpose:
 *	Updates the specific XLS with the contents of the ad-in's cache
 *	
 * Parameters
 *	WorkbookName	Name of the workbook to update
 *	WorksheetName	Name of the worksheet to update
 *	Range			Upper left cell of range to receive update
 *
 * Return Value:
 *  BOOL
 *
 */

STDMETHODIMP_(short) 
CDataObject::GetData(BSTR WorkbookName, BSTR WorksheetName, BSTR Range){
	LPTSTR prange = Range;
	TCHAR szCol[8];	// to hold a number < 1000000
	TCHAR szRow[8]; // to hold a number  < 1000000
	UINT uiCol = 0;
	UINT uiRow = 0;
	int i = 0;

	TRY{
		// Initialize strings
    	_fmemset(szCol, 0, sizeof(szCol));
    	_fmemset(szRow, 0, sizeof(szRow));
    
		// Parse string to get column 
		while((*prange >= (TCHAR) 65) && (*prange <= (TCHAR) 122)){
			szCol[i++] = *prange++;
			if(i >= 8){
				AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObject::GetData")
					, IDS_INVALIDEXCELCOLUMN
					, DISP_E_EXCEPTION); 

					MYRAISEEXCEPTION;
			}
		}
		
		if(lstrlen(szCol) > 1){
			uiCol = (((UINT)szCol[0] - 65) * 26) + ((UINT)szCol[1] - 65);		
		}else if(lstrlen(szCol) < 3){   
			if((UINT) szCol[0] > 96){
				uiCol = (UINT) szCol[0] - 97;
			}else{
				uiCol = (UINT) szCol[0] - 65;
			}
		}else{                   

			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObject::GetData")
					, IDS_DEBUG_COULDNOTCALCCOL
					, DISP_E_EXCEPTION); 

			MYRAISEEXCEPTION;
		}		      
		
		if(uiCol < 0 || uiCol > 256){

			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObject::GetData")
					, IDS_DEBUG_COULDNOTCALCCOL
					, DISP_E_EXCEPTION); 

			MYRAISEEXCEPTION;
		}

		// Parse string to get row
		lstrcpy(szRow, prange);
		
		#ifdef _WIN32
			uiRow = (UINT) _tcstol(szRow, NULL, 10) - 1;
		#else                   
			uiRow = (UINT) atol(szRow) - 1;
		#endif

        if(m_pRequests->m_col.m_cElements == 0){

			AddInDebugOutput(&m_excepinfo
					, TEXT("CDataObject::GetData")
					, IDS_DEBUG_COLLECTIONISEMPTY
					, DISP_E_EXCEPTION); 

			return TRUE;
        }                         
        
		// do we do a single cell update 
			                             
		if(m_pDirtyCells->IsSingleCellRefreshable()){	
			// if so, do an xlSet for each cell that changed
			
			for(int i=0;i<m_pDirtyCells->GetCount();i++){
				DWORD dwCell = m_pDirtyCells->GetCell(i);

				WORD x = LOWORD(dwCell);
				WORD y = HIWORD(dwCell);

				LPXLOPER pxl = &m_pCache->m_pxlCache[(x * (m_dwMaxProperties - m_bDontIncludeNames)) + y];
			
				if(FAILED(MyXLSet(pxl, WorkbookName, WorksheetName, uiRow + x, (BYTE) (uiCol + y), (UINT) 1, (UINT) 1))){

					AddInDebugOutput(&m_excepinfo
						, TEXT("CDataObject::GetData")
						, IDS_DEBUG_XLSETFAILED
						, DISP_E_EXCEPTION); 

				}
			}
			
			m_pDirtyCells->Reset();
		}else{
			// else if UpdateMaxrange was set to TRUE
			// update the xls's entire allocation
			if(m_bUpdateMaxRange && m_dwMaxRequests){

				if(FAILED(MyXLSet(m_pCache->m_pxlCache, WorkbookName
						, WorksheetName, uiRow, (BYTE) uiCol
						, (UINT) m_dwMaxProperties - m_bDontIncludeNames
						, (UINT) m_dwMaxRequests))){

					AddInDebugOutput(&m_excepinfo
						, TEXT("CDataObject::GetData")
						, IDS_DEBUG_XLSETFAILED
						, DISP_E_EXCEPTION); 

					MYRAISEEXCEPTION;
				}	

			}else{
				// else update xls with the exact size of the DataObject
				if(FAILED(MyXLSet(m_pCache->m_pxlCache, WorkbookName
						, WorksheetName, uiRow, (BYTE) uiCol
						, (UINT) m_dwMaxProperties - m_bDontIncludeNames
						, (UINT) m_pRequests->m_col.m_cElements))){

					AddInDebugOutput(&m_excepinfo
						, TEXT("CDataObject::GetData")
						, IDS_DEBUG_XLSETFAILED
						, DISP_E_EXCEPTION); 

					MYRAISEEXCEPTION;
				}
			}	
		}
		
		return TRUE;
	}                                   

	EXCEPT{
  	    m_bRaiseException = TRUE;
	    
 		AddInDebugOutput(&m_excepinfo, TEXT("CDataObject::GetData")
   			, IDS_DEBUG_EXCEPTION, NULL); 
	    
		return FALSE;
	}
}
                                                   
/*
 * CDataObject::OnDataChange		
 * Property - Read/Write	
 *
 * Purpose:
 *	Sets VBA function which will be called each time the data changes
 *
 * Return Value:
 *  BOOL
 *
 */

STDMETHODIMP_(void) 
CDataObject::put_OnDataChange(BSTR newvalue){
    m_bRaiseException = FALSE;

	if(FAILED(CheckPointer(newvalue))){
		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_OnDataChange")
			, IDS_DEBUG_BADPOINTER
			, DISP_E_EXCEPTION); 

	    m_bRaiseException = TRUE;
		    
	   	return;
	}

	if(m_pRealTimeExt->m_pXLAppDispatch == NULL){
		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_OnDataChange")
			, IDS_DEBUG_XLAPPLICATIONPROPNOTSET
			, DISP_E_EXCEPTION); 

	    m_bRaiseException = TRUE;
		    
	   	return;
	}

	if(m_bstrXLCallBackFunc){
		SysFreeString(m_bstrXLCallBackFunc);
		m_bstrXLCallBackFunc = NULL;
	}

	m_bstrXLCallBackFunc = SysAllocString(newvalue);

	if(m_bstrXLCallBackFunc == NULL){
		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_OnDataChange")
			, IDS_DEBUG_BSTRREALLOCFAILED
			, DISP_E_EXCEPTION); 

	    m_bRaiseException = TRUE;
		    
	   	return;
	}

	return;		
}

STDMETHODIMP_(BSTR) 
CDataObject::get_OnDataChange(){
	return m_bstrXLCallBackFunc;	                         
}


/*
 * CDataObject::MaxRequests
 * Property - Read/Write before Requests::Add
 *
 * Purpose:
 *	Maximum number of requests DataObject can contain, If zero, requests
 *	will be realloced dynamically with no limitation.
 *
 * Return Value:
 *  BOOL
 *
 */

STDMETHODIMP_(void)
CDataObject::put_MaxRequests(long Count){
	if(m_pRequests->m_col.m_cElements > 0){
		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_MaxRequests")
			, IDS_DEBUG_PROPERTYISREADONLYAFTERREQUESTSADD
			, DISP_E_EXCEPTION); 

	    m_bRaiseException = TRUE;

		return;
	}

	// Cannot be larger than 32000
	if(Count > MAX_REQUESTS || Count < 0){

		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_MaxRequests")
			, IDS_DEBUG_INVALIDPROPVALUE
			, DISP_E_EXCEPTION); 

	    m_bRaiseException = TRUE;

	   	return;
	}

	m_dwMaxRequests = Count;
	
	return;
}

STDMETHODIMP_(long)
CDataObject::get_MaxRequests(){
	return (long) m_dwMaxRequests;
}

/*
 * CDataObject::MaxProperties
 * Property - Read/Write before Requests::Add
 *
 * Purpose:
 *	Maximum number of properties a Request can contain.
 *
 * Return Value:
 *  BOOL
 *
 */

STDMETHODIMP_(void)
CDataObject::put_MaxProperties(long Count){
    m_bRaiseException = FALSE;

	if(m_pRequests->m_col.m_cElements > 0){
		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_MaxProperties")
			, IDS_DEBUG_PROPERTYISREADONLYAFTERREQUESTSADD
			, DISP_E_EXCEPTION); 

	    m_bRaiseException = TRUE;

		return;
	}
	
	// Cannot be greater than max value of WORD
	if(Count > MAX_PROPERTIES || Count < 1){
		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_MaxProperties")
			, IDS_DEBUG_INVALIDPROPVALUE
			, DISP_E_EXCEPTION); 

	    m_bRaiseException = TRUE;

		return;
	}

	m_dwMaxProperties = Count;
		
	return;
}

STDMETHODIMP_(long)
CDataObject::get_MaxProperties(){
	return m_dwMaxProperties;
}

/*
 * CDataObject::IncludeRequestNames
 * Property - Read/Write before Requests::Add
 *
 * Purpose:
 *	When GetData is called should add-in include Request Names when
 *	updating xls
 *
 * Return Value:
 *  BOOL
 *
 */

STDMETHODIMP_(void)
CDataObject::put_IncludeRequestNames(short Value){
	if(m_pRequests->m_col.m_cElements > 0){

		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_IncludeRequestNames")
			, IDS_DEBUG_PROPERTYISREADONLYAFTERREQUESTSADD
			, DISP_E_EXCEPTION); 

	    m_bRaiseException = TRUE;

		return;
	}else{
		if(Value){
			m_bDontIncludeNames = 0;
		}else{
			m_bDontIncludeNames = 1;
		}
	}
		
	return;
}

STDMETHODIMP_(short)
CDataObject::get_IncludeRequestNames(){
	return !m_bDontIncludeNames;
}


/*
 * CDataObject::UpdateMaxRange
 * Property - Read/Write
 *
 * Purpose:
 *	This property is valid only if MaxRequests is set to a valid number.
 *  When GetData is called the xls range that is updated will be as large
 *	as MaxRequests value, blanking out final requests which do not have values.
 *	This is used for DataObject updating where the definition of the DataObject
 *	reduces.
 *
 * Return Value:
 *  BOOL
 *
 */

STDMETHODIMP_(void)
CDataObject::put_UpdateMaxRange(short Value){
	m_bUpdateMaxRange = Value;
}

STDMETHODIMP_(short)
CDataObject::get_UpdateMaxRange(){
	return	m_bUpdateMaxRange;
}

/*
 * CDataObject::UpdatePerCellCount
 * Property - Read/Write
 *
 * Purpose:
 * 	Normally, if the add-in will update the entire DataObject. If the number
 * 	of cell changes per update is less than UpdatePerCellCount, the add-in will
 * 	do an xlSet for each.
 * 
 * Return Value:
 *  BOOL
 *
 */

STDMETHODIMP_(void)
CDataObject::put_UpdatePerCellCount(short Value){
	if(Value < DIRTY_CELL_MAX){
		m_pDirtyCells->SetMaxCount(Value);
	}else{
		AddInDebugOutput(&m_excepinfo
			, TEXT("CDataObject::put_UpdatePerCellCount")
			, IDS_DEBUG_INVALIDPROPVALUE
			, DISP_E_EXCEPTION); 
                   
	    m_bRaiseException = TRUE;
                   
		return;
	}
}

STDMETHODIMP_(short)
CDataObject::get_UpdatePerCellCount(){
	return	m_pDirtyCells->GetMaxCount();
}

                                                   
/************************************************
* 		Heper DataObject Dispatchs
*
************************************************/

/*
 * CDataObject::XRTRequests
 *
 * Purpose:
 *	Asks WOSA/XRT DataObject for its Requests
 *
 * Parameter
 *	ppDispatch 		Pointer to return XRT's ppDispatch
 * 
 * Return Value:
 *  HRESULT
 *
 */
    
HRESULT CDataObject::XRTRequests(LPDISPATCH FAR *ppDispatch){
	static DISPID dispid = -1;
	static DISPPARAMS dispparams = {NULL, NULL, 0, 0};
	static TCHAR szMember[] = TEXT("Requests");
	VARIANT vResult;  

	TRY{	
		if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::XRTRequests")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppDispatch = NULL;
		
		VariantInit(&vResult);
	
		if(FAILED(InvokePropGetHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){
			return ResultFromScode(E_FAIL);
		}

		if(FAILED(CheckVariant(&vResult, TRUE))){
			// If its a bad ptr I cant call release
			MYRAISEEXCEPTION;
		}

		*ppDispatch = vResult.pdispVal;

		// I am not sure if you need this addref.
		// Your WOSA/XRT object should do it for you

		(*ppDispatch)->AddRef();

		return S_OK;
	}
	
	EXCEPT{                               
		VariantClear(&vResult);
		return ResultFromScode(E_FAIL);
	}
}

/*
 * CDataObject::XRTDataItems
 *
 * Purpose:
 *	Asks WOSA/XRT DataObject for its DataItems
 *
 * Parameter
 *	ppDispatch 		Pointer to return XRT's ppDispatch
 * 
 * Return Value:
 *  HRESULT
 *
 */

HRESULT CDataObject::XRTDataItems(LPDISPATCH FAR *ppDispatch){
	static DISPID dispid = -1;
	static DISPPARAMS dispparams = {NULL, NULL, 0, 0};
	static TCHAR szMember[] = TEXT("DataItems");
	VARIANT vResult;

	TRY{	
		if(FAILED(CheckPointer((LPVOID) ppDispatch, 0))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::XRTDataItems")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*ppDispatch = NULL;
	
		VariantInit(&vResult);
	
		if(FAILED(InvokePropGetHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, &vResult))){
			return ResultFromScode(E_FAIL);
		}
		
		if(FAILED(CheckVariant(&vResult, TRUE))){
			// If its a bad ptr I cant call release
			MYRAISEEXCEPTION;
		}

		*ppDispatch = vResult.pdispVal;

		// I am not sure if you need this addref.
		// Your WOSA/XRT object should do it for you
		(*ppDispatch)->AddRef();
	
		return S_OK;
	}
	
	EXCEPT{       
		VariantClear(&vResult);
		return ResultFromScode(E_FAIL);
	}
}


/*
 * CDataObject::RetrieveDataObjectName
 * Property - Read/Write
 *
 * Purpose:
 *	asks WOSA/XRT DataObject for its Name
 * 
 * Return Value:
 *  HRESULT
 *
 */

HRESULT CDataObject::RetrieveDataObjectName(LPDISPATCH pDispatch
		, BSTR FAR *pbstrName){
	static DISPID dispid = -1;
	static DISPPARAMS dispparams = {NULL, NULL, 0, 0};
	static TCHAR szMember[] = TEXT("Name");
	VARIANT vResult;

	TRY{
		if(FAILED(CheckPointer((LPVOID) pbstrName, 0))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::RetreiveDataObjectName")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}
		
		*pbstrName = NULL;

		VariantInit(&vResult);
	
		if(FAILED(InvokePropGetHelper(pDispatch, &dispid, 
			&dispparams, szMember, &vResult))){

   			AddInDebugOutput(NULL, TEXT("CDataObject::SetDataObjectName")
   				, IDS_CALLTOXRTDATAOBJNAMEFAILED
   				, NULL); 
			MYRAISEEXCEPTION;
		}
        
        if(FAILED(CheckVariant(&vResult, TRUE))){
   			AddInDebugOutput(NULL, TEXT("CDataObject::RetreiveDataObjectName")
   				, IDS_DEBUG_BADVARIANT, NULL); 
			MYRAISEEXCEPTION;
        }

		*pbstrName = vResult.bstrVal;
	}
	
	EXCEPT{
		VariantClear(&vResult);

		return ResultFromScode(E_FAIL);	
	}
}
    

HRESULT CDataObject::SetDataObjectName(BSTR bstrName){
	static DISPID dispid = -1;
	static TCHAR szMember[] = TEXT("Name");
	DISPPARAMS dispparams = {NULL, NULL, 0, 0};
	VARIANT vParam;
	DISPID dispidNamedVars = DISPID_PROPERTYPUT;
	
	TRY{
		if(FAILED(CheckPointer((LPVOID) bstrName, 0))){
    		AddInDebugOutput(NULL, TEXT("CDataObject::SendDataObjectName")
    			, IDS_DEBUG_BADPOINTER, NULL); 
			return ResultFromScode(E_FAIL);
		}

		VariantInit(&vParam);
		vParam.vt = VT_BSTR;
		vParam.bstrVal = SysAllocString(bstrName);

		dispparams.rgvarg = &vParam;
		dispparams.rgdispidNamedArgs = &dispidNamedVars;
		dispparams.cArgs = 1;
		dispparams.cNamedArgs = 1;
	
		if(FAILED(InvokePropPutHelper(m_pXRTDispatch, &dispid, 
				&dispparams, szMember, NULL))){

   			AddInDebugOutput(NULL, TEXT("CDataObject::SetDataObjectName")
   				, IDS_CALLTOXRTDATAOBJNAMEFAILED
   				, NULL); 
			MYRAISEEXCEPTION;
		}

		VariantClear(&vParam);
	}
	
	EXCEPT{
		VariantClear(&vParam);

		return ResultFromScode(E_FAIL);	
	}
}
