// 	WOSA/XRT Excel 5.0 AddIn Version 1.00
//
//	COPYRIGHT:
//
//  	(C) Copyright Microsoft Corp. 1994.  All rights reserved.
//
//   	You have a royalty-free right to use, modify, reproduce and
//   	distribute the Sample Files (and/or any modified version) in
//   	any way you find useful, provided that you agree that
//   	Microsoft has no warranty obligations or liability for any
//   	Sample Application Files which are modified.
//
//
// 	Revision History
//    	October 15, 1994   rc    Version 1.00 Release
//
// 	Comments:
//		Initialization module

#define GLOBAL
#include "precomp.h"
#include "framewrk.h"
#include "fn.h"
#include "cache.h"

#include "xrtxll.h"

#include "realtime.h"
#include "dataobj.h"
#include "dataobjs.h"
#include "xrtcf.h"

#include "xrtxlreg.h"

#include "guid.h"

extern "C"
void EXPORT FAR PASCAL xlAutoFree(LPXLOPER px);

void InitializeAddIn(HWND hWnd);

// rgFuncs
//
// This is a table of all the functions exported by this module.
// These functions are all registered (in xlAutoOpen) when you
// open the XLL. Before every string, leave a space for the
// byte count. The format of this table is the same as 
// the last seven arguments to the REGISTER function.
// rgFuncsRows define the number of rows in the table. The
// dimension [7] represents the number of columns in the table.
//
static LPSTR rgFuncs[][7] =
{                                      
    
   	 {" REALTIMEUPDATE"," RRRR", " REALTIMEUPDATE", " DataObject, Item, Property", " 1", " REALTIMEUPDATE", " "}
};       

#define rgFuncsRows (sizeof(rgFuncs)/sizeof(rgFuncs[0]))

//
// rgMenu
//
// This is a table describing the XRTXLL drop-down menu. It is in
// the same format as the Microsoft Excel macro language menu tables.
// The first column contains the name of the menu or command, the
// second column contains the function to be executed, the third
// column contains the (Macintosh only) shortcut key, the fourth
// column contains the help text for the status bar, and
// the fifth column contains the help text index. Leave a space
// before every string so the byte count can be inserted. rgMenuRows
// defines the number of menu items. 5 represents the number of
// columns in the table.
//

static LPSTR rgMenu[][5] =
{
    {" ",          " ",           " ",
    " ",         " "}
};

#define rgMenuRows (sizeof(rgMenu)/sizeof(rgMenu[0]))

#ifdef WIN32
extern "C"
int WINAPI DllEntryPoint(HANDLE hInstance, DWORD dw , LPVOID lpvReserved)
{

#ifdef _DEBUG
    OutputDebugString( TEXT("XRTXLL32:  DllMain\r\n") ) ;
#endif

    if (dw == DLL_PROCESS_ATTACH)
        return LibMain( hInstance, 0, 0, NULL ) ;
    else if (dw == DLL_PROCESS_DETACH)
        return WEP ( dw ) ;
    else
        return 0 ;
}
#endif

// LibMain
//
// This function is called by LibEntry (in LIBENTRY.OBJ), which is called
// by Windows when the DLL is first loaded. LibEntry initializes the
// DLL's heap if a HEAPSIZE is specified in the DLL's .DEF file, and
// then calls LibMain. The following LibMain function satisfies that
// call. The LibMain function should perform additional initialization
// tasks required by the DLL. In this example, we byte-count all the strings
// in the preceding table. LibMain will be called only once, when a DLL is
// first loaded, even if that DLL is used by multiple applications.
//
// LibMain
//
// Arguments:
//
//      HANDLE hInstance    Instance handle
//      WORD wDataSeg       Library's data segment
//      WORD wHeapSize      Heap size
//      LPTSTR lpszCmdLine   Command line
//
// Returns:
//
//      int                 1 if initialization is successful.

extern "C"
int EXPORT FAR PASCAL LibMain(HANDLE hInstance, UINT wDataSeg,
       UINT wHeapSize, LPTSTR lpszCmdLine)
{
  	g_hinst = hInstance;
		
    return 1;
}

// WEP
//
// This is the DLL's termination function, which is called when the
// DLL terminates. This function should return 1. Microsoft
// QuickC for Windows provides a built-in WEP that eliminates the
// need for this routine.
//
// Arguments:
//
//      int nArgument      WEP_SYSTEMEXIT if Windows is
//                         shutting down, or
//                         WEP_FREE_DLL if this DLL is
//                         being unloaded by Windows.
//
// Returns:
//
//      int                1 (required)


extern "C"
int EXPORT FAR PASCAL WEP (int nArgument)
{
    #ifdef _DEBUG
    OutputDebugString(TEXT("XRTXLL: WEP\r\n")) ;
    #endif
	
	if(g_pRealTimeExt)
    	g_pRealTimeExt->Release();
          
    return 1;
}

// xlAutoOpen
//
// Microsoft Excel uses xlAutoOpen to load XLL files.
// When you open an XLL file, the only action
// Microsoft Excel takes is to call the xlAutoOpen function.
//
// More specifically, xlAutoOpen is called:
//
//  - when you open this XLL file from the File menu,
//  - when this XLL is in the XLSTART directory, and is
//    automatically opened when Microsoft Excel starts,
//  - when Microsoft Excel opens this XLL for any other reason, or
//  - when a macro calls REGISTER(), with only one argument, which is the
//    name of this XLL.
//
// xlAutoOpen is also called by the Add-in Manager when you add this XLL 
// as an add-in. The Add-in Manager first calls xlAutoAdd, then calls
// REGISTER("EXAMPLE.XLL"), which in turn calls xlAutoOpen.
//
// xlAutoOpen should:
//
//  - register all the functions you want to make available while this
//    XLL is open,
//
//  - add any menus or menu items that this XLL supports,
//
//  - perform any other initialization you need, and
//
//  - return 1 if successful, or return 0 if your XLL cannot be opened.
//
extern "C"
int EXPORT FAR PASCAL xlAutoOpen(void)
{

    static XLOPER   xDLL;             	// name of this DLL //

    int i=0,j=0;

    Excel(xlGetName, &xDLL, 0);

    for (i=0; i<rgFuncsRows; i++)
    {
        for (j=0; j<7; j++)
        {
            rgFuncs[i][j][0] = (BYTE)lstrlenA(rgFuncs[i][j]+1);
        }
    }

    for (i=0; i<rgFuncsRows; i++){
	    Excel(xlfRegister, NULL, 8,
    	    (LPXLOPER) &xDLL,
        	(LPXLOPER) TempStr(rgFuncs[i][0]),
	        (LPXLOPER) TempStr(rgFuncs[i][1]),
    	    (LPXLOPER) TempStr(rgFuncs[i][2]),
        	(LPXLOPER) TempStr(rgFuncs[i][3]),
	        (LPXLOPER) TempStr(rgFuncs[i][4]),
    	    (LPXLOPER) TempStr(rgFuncs[i][5]),
        	(LPXLOPER) TempStr(rgFuncs[i][6]));
    }
    
    Excel( xlGetHwnd, &xDLL, 0 ) ;
   	HWND hwndXL = (HWND)xDLL.val.w ;

	InitializeAddIn(hwndXL);

    return 1;
}
		

//
// xlAutoClose
//
// xlAutoClose is called by Microsoft Excel:
//
//  - when you quit Microsoft Excel, or
//  - when a macro sheet calls UNREGISTER(), giving a string argument
//    which is the name of this XLL.
//
// xlAutoClose is called by the Add-in Manager when you remove this XLL from
// the list of loaded add-ins. The Add-in Manager first calls xlAutoRemove,
// then calls UNREGISTER("EXAMPLE.XLL"), which in turn calls xlAutoClose.
//
// xlAutoClose is called by EXAMPLE.XLL by the function fExit. This function
// is called when you exit Example.
//
// xlAutoClose should:
//
//  - Remove any menus or menu items that were added in xlAutoOpen,
//
//  - do any necessary global cleanup, and
//
//  - delete any names that were added (names of exported functions, and 
//    so on). Remember that registering functions may cause names to be created.
//
// xlAutoClose does NOT have to unregister the functions that were registered
// in xlAutoOpen. This is done automatically by Microsoft Excel after
// xlAutoClose returns.
//
// xlAutoClose should return 1.
//
extern "C"
int EXPORT FAR PASCAL xlAutoClose(void)
{
    int i;
    XLOPER xRes;

    // This block first deletes all names added by xlAutoOpen or
    // xlAutoRegister. Next, it checks if the drop-down menu XRTXLL still
    // exists. If it does, it is deleted using xlfDeleteMenu. 
    //

    for (i=0; i<rgFuncsRows; i++)
    {
        Excel(xlfSetName, 0, 1, TempStr(rgFuncs[i][2]));
    }

    Excel(xlfGetBar, &xRes, 3, TempInt(1), TempStr(" Real-time"), TempInt(0));

    if (xRes.xltype != xltypeErr)
    {
        Excel(xlfDeleteMenu, 0, 2, TempNum(1), TempStr(" Real-time"));
    
        // Free the XLOPER returned by xlfGetBar //
        Excel(xlFree, 0, 1, (LPXLOPER) &xRes);
    }	  


    return 1;
}


// xlAutoRegister
//
// This function is called by Microsoft Excel if a macro sheet tries to
// register a function without specifying the type_text argument. If that
// happens, Microsoft Excel calls xlAutoRegister, passing the name of the
// function that the user tried to register. xlAutoRegister should use the
// normal REGISTER function to register the function, only this time it must
// specify the type_text argument. If xlAutoRegister does not recognize the
// function name, it should return a #VALUE! ResultFromScode(E_FAIL). Otherwise, it should
// return whatever REGISTER returned.
//
// Arguments:
//
//      LPXLOPER pxName     xltypeStr containing the
//                          name of the function
//                          to be registered. This is not
//                          case sensitive.
//
// Returns:
//
//      LPXLOPER            xltypeNum containing the result
//                          of registering the function,
//                          or xltypeErr containing #VALUE!
//                          if the function could not be
//                          registered.
//
extern "C"
LPXLOPER EXPORT FAR PASCAL xlAutoRegister(LPXLOPER pxName)
{
    static XLOPER xDLL, xRegId;
    int i;

    //
    // This block initializes xRegId to a #VALUE! ResultFromScode(E_FAIL) first. This is done in
    // case a function is not found to register. Next, the code loops through 
    // the functions in rgFuncs[] and uses LPTSTRicmp to determine if the 
    // current row in rgFuncs[] represents the function that needs to be 
    // registered. When it finds the proper row, the function is registered 
    // and the register ID is returned to Microsoft Excel. If no matching 
    // function is found, an xRegId is returned containing a #VALUE! ResultFromScode(E_FAIL).
    //

    xRegId.xltype = xltypeErr;
    xRegId.val.err = xlerrValue;

    for (i=0; i<rgFuncsRows; i++)
    {
        if (!lstrcmpiA(rgFuncs[i][0], pxName->val.str))
        {
            Excel(xlGetName, &xDLL, 0);
    
            Excel(xlfRegister, &xRegId, 8,
            (LPXLOPER) &xDLL,
            (LPXLOPER) TempStr(rgFuncs[i][0]),
            (LPXLOPER) TempStr(rgFuncs[i][1]),
            (LPXLOPER) TempStr(rgFuncs[i][2]),
            (LPXLOPER) TempStr(rgFuncs[i][3]),
            (LPXLOPER) TempStr(rgFuncs[i][4]),
            (LPXLOPER) TempStr(rgFuncs[i][5]),
            (LPXLOPER) TempStr(rgFuncs[i][6]));
    
            //* Free oper returned by xl //
            Excel(xlFree, 0, 1, (LPXLOPER) &xDLL);
    
            return (LPXLOPER) &xRegId;
        }
    }
    return (LPXLOPER) &xRegId;
}

// xlAutoAdd
//
// This function is called by the Add-in Manager only. When you add a
// DLL to the list of active add-ins, the Add-in Manager calls xlAutoAdd()
// and then opens the XLL, which in turn calls xlAutoOpen.
//
//
extern "C"
int EXPORT FAR PASCAL xlAutoAdd(void)
{
    // Display a dialog box indicating that the XLL was successfully added //
    Excel(xlcAlert, 0, 2, TempStr(" Real-Time Extensions Installed."),
      TempInt(2));
    return 1;
}


//
// xlAutoRemove
//
// This function is called by the Add-in Manager only. When you remove
// an XLL from the list of active add-ins, the Add-in Manager calls
// xlAutoRemove() and then UNREGISTER("EXAMPLE.XLL").
//
// You can use this function to perform any special tasks that need to be
// performed when you remove the XLL from the Add-in Manager's list
// of active add-ins. For example, you may want to delete an
// initialization file when the XLL is removed from the list.
//
extern "C"
int EXPORT FAR PASCAL xlAutoRemove(void)
{
    // Show a dialog box indicating that the XLL was successfully removed //
    Excel(xlcAlert, 0, 2, TempStr(" Real-Time Extensions Removed."),
      TempInt(2));
    return 1;
}

//
// xlAddInManagerInfo
//
//
// This function is called by the Add-in Manager to find the long name
// of the add-in. If xAction = 1, this function should return a string
// containing the long name of this XLL, which the Add-in Manager will use
// to describe this XLL. If xAction = 2 or 3, this function should return
// #VALUE!.
//
// Arguments
//
//      LPXLOPER xAction    What information you want. One of:
//                          1 = the long name of the
//                              add-in
//                          2 = reserved
//                          3 = reserved
// Return value
//
//      LPXLOPER            The long name or #VALUE!.
//
//
extern "C"
LPXLOPER EXPORT FAR PASCAL xlAddInManagerInfo(LPXLOPER xAction)
{
    static XLOPER xInfo, xIntAction;

    //
    // This code coerces the passed-in value to an integer. This is how the
    // code determines what is being requested. If it receives a 1, 
    // it returns a string representing the long name. If it receives 
    // anything else, it returns a #VALUE! ResultFromScode(E_FAIL).
    //

    Excel(xlCoerce, &xIntAction, 2, xAction, TempInt(xltypeInt));

    if(xIntAction.val.w==1)
    {
        xInfo.xltype = xltypeStr;
        xInfo.val.str = "\026Real-Time Extensions";
    }
    else
    {
        xInfo.xltype = xltypeErr;
        xInfo.val.err = xlerrValue;
    }

    return (LPXLOPER) &xInfo;
}

// xlAutoFree
//
extern "C"
void EXPORT FAR PASCAL xlAutoFree(LPXLOPER px)
{
    return;
}

BOOL InitWindow(HINSTANCE hInst){
	WNDCLASS wc;

	wc.style = CS_DBLCLKS;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 12;
	wc.hInstance = hInst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH) GetStockObject(GRAY_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = TEXT("XRTXLL_CLB_RC");
	
	if(!RegisterClass(&wc))
		return TRUE;
		
	return FALSE;
}
	
LONG EXPORT FAR PASCAL WndProc(HWND hWnd, UINT msg, WPARAM wp, LPARAM lp){
	#ifdef DEBUG
		TCHAR szDebug[20] = {TEXT("")};
	#endif
	
	TRY{
		if(msg == g_msgXRTDBUpdate){
			static HRESULT hr;
			static CDataObject FAR *pDataObject = NULL;
		
	    	pDataObject = (CDataObject FAR *)GetWindowLong(hWnd, 8);
        
        	if(pDataObject == NULL){
				AddInDebugOutput(NULL
					, TEXT("IAdviseSink::OnDataChange")
					, IDS_DEBUG_INAVALIDCDATAOBJPTR
					, NULL); 
				return 1L;
	        }

		    LPSTGMEDIUM lpStg = (LPSTGMEDIUM) lp;
        
			if(FAILED(pDataObject->m_pCache->Update(lpStg))){
				AddInDebugOutput(NULL 
					, TEXT("StatusWndProc")
					, IDS_DEBUG_CACHEUPDATEFAILED
					, NULL); 
				MYRAISEEXCEPTION;
			}                       
			
			if(pDataObject->m_pUpdateRangeDispatch){
				XLCalculateNow(pDataObject);
			}
			
			if(pDataObject->m_bstrXLCallBackFunc != NULL) {
				XLRunMacro(pDataObject);
			}
		}else if(msg == WM_PAINT){
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			RECT cr;
			
			GetClientRect(hWnd, &cr);
			
			SetBkMode(hdc, TRANSPARENT);

			TextOut(hdc, 10 * g_nBorderX, 5 * g_nBorderY
				, (LPTSTR) g_pRealTimeExt->m_pDataObjects->m_bstrStatusWndMsg
				, SysStringLen(g_pRealTimeExt->m_pDataObjects->m_bstrStatusWndMsg));
				
			HBRUSH hb = CreateSolidBrush(RGB(0,0,255));

			cr.top = (int) (cr.bottom - (15 * g_nBorderY));
			cr.bottom = cr.top + (10 * g_nBorderY);

			cr.left = (10 * g_nBorderX);
			int nMaxBarWidth = cr.right - (2 * cr.left);

			cr.right = cr.left + (int) (g_pRealTimeExt->m_pDataObjects->m_fComplete * (float) nMaxBarWidth);

			FillRect(hdc, &cr, hb);
			
			DeleteObject(hb);
			
			EndPaint(hWnd, &ps);		
		}
	
		return DefWindowProc(hWnd, msg, wp, lp);
	}

	EXCEPT{                  
   		AddInDebugOutput(NULL, TEXT("WndProc")
   			, IDS_DEBUG_EXCEPTION, NULL); 
		return 0L;
	}
}	
	
HRESULT EXPORT FAR PASCAL DllGetClassObject(REFCLSID rclsid, REFIID riid
    , LPVOID FAR *ppv)
{       
	LPCLASSFACTORY pXrtCF = NULL;
	
	TRY{
		if(FAILED(CheckPointer((LPVOID) ppv, 0))){
			MYRAISEEXCEPTION;
		}

	    *ppv=NULL;

		XLOPER xDLL;

	    Excel( xlGetHwnd, &xDLL, 0 ) ;
    	HWND hwndXL = (HWND)xDLL.val.w ;

    	if(hwndXL == 0){
     		MessageBox(NULL, TEXT("Add-In must be installed into Excel 5.0 before accessing."), g_szTitle, MB_OK);       
        	return ResultFromScode(E_FAIL);
	    }
		        
		pXrtCF = new CXrtCF();
    
    	if (pXrtCF == NULL){
    		MYRAISEEXCEPTION;
   	 	}
    
	    *ppv = pXrtCF;
		return S_OK;    
	}

	EXCEPT{
	    if (pXrtCF)   
    	    pXrtCF->Release();
    
    	return ResultFromScode(E_FAIL);
	}
}

STDAPI DllCanUnloadNow(void)
{
	if(g_cObjCount){
		return ResultFromScode(ResultFromScode(E_FAIL));		
	}else{
		return S_OK;		
	}
}


void InitializeAddIn(HWND hWnd){
	#ifdef _WIN32
   		g_cfWOSAXRT = RegisterClipboardFormatA(szCF_WOSAXRT101A);      
   	#else
   		g_cfWOSAXRT = RegisterClipboardFormat(szCF_WOSAXRT101A);      
   	#endif
	
	g_msgXRTDBUpdate = RegisterWindowMessage((LPTSTR)TEXT("XRTXLL DB Message Update"));

	XLOPER xRes;
	Excel(xlGetInst, (LPXLOPER) &xRes, 0);
	HINSTANCE hInstance = (HINSTANCE) xRes.val.w;

	InitWindow(hInstance);    

	#ifdef WIN32
		HANDLE hMod = GetModuleHandle(FILENAME);
	#else
		HANDLE hMod = g_hinst;	
	#endif
	
	if(NULL == LoadString(hMod, IDS_TITLE, g_szTitle, sizeof(g_szTitle))){
		lstrcpy(g_szTitle, TEXT("Internal error. Cannot provide desc of error."));
	}
	
	OleRegisterServerClass(
		CLSID_App
		, TEXT("Excel Real-Time Extensions")
		, TEXT("Realtime.Extensions.1")
		, TEXT("Realtime.Extensions")
		, LIBID_App
		, NULL
		, NULL );
}           	
